#include <iostream>
#include "mpreal.h"
#include <vector>
#include <random>
#include "CMA.hpp"
#include <cmaes.h>
//#include <math.h>
#include <cmath>


using namespace std;
using namespace mpfr;
using namespace libcmaes;

class HyperPoly {
public:
    vector<mpreal> a;
   
    HyperPoly(int n) {
        a.resize(n);
    }
    void fromPolySquare (const vector<mpreal>& phi);
    mpreal cv2();
};

int Nmax=10000;
mpreal mpfact[20008];//2*Nmax+3
static int precision=33300; //Nmax*3.32  binry precision
static mpreal mpZero (0,precision);
static mpreal mpOne (1,precision);
static mpreal mpTwo (2,precision);

void compPolyCoeffs(const vector<mpreal>& phi,
                HyperPoly& target) {

    target.a.resize(2*phi.size()+1);
    target.a[0]= mpOne ;
    mpreal temp;

    for (int i=0; i<phi.size(); i++) {
	for (int k=2*phi.size(); k>=0; k--) {
          temp = mpZero;
          if(k>1) temp = target.a[k-2];
          if(k>0) temp -= mpTwo * phi[i] * target.a[k-1];
          target.a[k] = temp +  phi[i] * phi[i] * target.a[k];
        }
    }
}

void HyperPoly::fromPolySquare(const vector<mpreal>& phi) {
    HyperPoly buffer(200);
    compPolyCoeffs(phi, *this);
}

static int globN;


mpreal HyperPoly::cv2() {

    mpreal m0 = mpZero;
    mpreal m1 = mpZero;
    mpreal m2 = mpZero;
    for (int i=0; i<a.size(); i++) {
        m0 += a[i] * mpfact[i]/mpfact[2*globN+1];
        m1 += a[i] * mpfact[i+1]/mpfact[2*globN+1];
        m2 += a[i] * mpfact[i+2]/mpfact[2*globN+1];
    }
    return m2*m0/m1/m1-mpOne;
}


mpreal cv2 (vector<mpreal> u) {

    HyperPoly target(u.size());

    vector<mpreal> phi (u.begin(), u.end());
    target.fromPolySquare(phi);
    mpreal cv2 = target.cv2();
    return cv2;
}

double funWithGap(double omega, double place, double gap, int N) {
    double delta = (2.0*M_PI-gap) / N;
    int i = floor(place/delta+0.5);
    vector<mpreal> params;
    params.push_back(omega);
    for (int j=1; j<=i; j++)
        params.push_back((j-0.5)*delta-M_PI);
    for (int j=i+1; j<=N; j++)
        params.push_back((j-0.5)*delta+gap-M_PI);
    return cv2(params).toDouble();
}


//=================================================================
double gap_cma_objective(const double* par) {
    mpreal::set_default_prec(mpfr::digits2bits(precision));
    //ascale, place, gap, pmax, istar, Ascale
    double ascale = abs(par[0]);
    double place = abs(par[1]);
    double gap = abs(par[2]);

    double pmax = abs(par[3]);
//    int istar = floor(abs(par[4]));
    int istar = std::floor(0.76*globN) ;
//    double Ascale = abs(par[5]);
    double Ascale = abs(par[4]);


    if(ascale>50.0) return 1.0+abs(ascale);
    if(Ascale>50.0) return 1.0+abs(Ascale);

//    if(gap<0.0) return 1.0+fabs(gap);
//    if(gap>50.0) return 1.0+abs(gap);


//    if(istar<globN/2.0) return 1.0+fabs(par[4]);
//    if(istar>std::round(0.99*globN)) return 1.0+fabs(par[4]);

    double pmin = 1.0/globN; 

    if(place-gap-pmin<0) return 1.0+abs(place-gap);

    if(place+gap>pmax) return 1.0+abs(place+gap);

    double buff;
 
//    double pmax = 3.8 * globN; 
//    double Ascale = 4.0;
//    int istar =  globN-floor(std::sqrt(globN/1.0));
    double pw = 1.0/ascale;

//    cout << "pmin=" << pmin <<endl ;
//    cout << "pmax=" << pmax <<endl ;
//    cout << "pw=" << pw <<endl ;
    
    buff = std::pow(pmin,pw);
 //   cout << "pmin^pw=" << buff <<endl ;
    
    double lambda = istar * std::pow(pmin,pw) - std::pow(place-gap,pw) ;

    
    lambda = lambda / ( std::pow(pmin,pw) - std::pow(place-gap,pw) ) ; 

//    cout << "lambda=" << lambda<<endl ;
    
    buff = (1.0 - istar  ) * std::pow(pmin,pw) / ( std::pow(pmin,pw) - std::pow(place-gap,pw) ) ;
    double tau = pmin * std::pow(buff, - ascale) ; 

//    cout << "tau=" << tau <<endl;
    
    vector<mpreal> params;
    for (int j=1; j<=istar; j++)
        params.push_back( tau * std::pow( j - lambda , ascale) );
 
 
    if(istar<globN-1){
    buff= std::pow(istar+1,Ascale)-std::pow(globN,Ascale);
    double Lambda = log((place+gap)/pmax)/buff;
    
    buff= std::pow(istar+1,Ascale)*log((place+gap)/pmax)/buff;
    double Theta = buff - log((place+gap));
    for (int j=istar+1; j<=globN; j++)
            params.push_back( exp(Lambda * std::pow(j, Ascale) - Theta )  );
    }
    if(istar==globN-1){
    params.push_back(pmax);
    }
    
/*    cout << "phi={" ;
     for (int i=0; i<globN; i++) {
        cout << params[i];
        if (i==globN-1)
            cout << "}" << endl;
        else
            cout << ", ";
    }
*/    
 
    return cv2(params).toDouble();
}

mpreal cma_objective_mp(const mpreal* par) {
    vector<mpreal> params(globN+1);
    for (int i=0; i<globN; i++)
        params[i] = par[i];
    return cv2(params);
}

double cma_objective(const double* par) {
    mpreal::set_default_prec(mpfr::digits2bits(precision));
    vector<mpreal> params(globN);
    for (int i=0; i<globN; i++)
        params[i] = mpreal(par[i],precision);
    return cv2(params).toDouble();
}

static FitFunc bipop_cma_objective = [](const double *par, const int N) {
    return cma_objective(par);
};

//=====================================================================
void gap_cmaes() {

//    double init[6] = {2.5, 2.5 * globN, 10.0, 3.8 * globN, std::floor(0.7*globN), 3.0}; 
    //ascale, place, gap, pmax, istar, Ascale

    double init[5] = {2.5, 2.5 * globN, 10.0, 3.8 * globN, 3.0}; 
    //ascale, place, gap, pmax, Ascale

    
    double cv = gap_cma_objective(init);
    int istar = std::floor(0.76*globN) ;
    cout << "init cv=" << cv << ", {n,a,p,w,max,istar,A}={" << globN << ", " << init[0] << ", " << init[1] << ", " << init[2] << ", " << init[3] << ", " << istar << ", " << init[4]  << "}" << endl;

//    CMA<double,double> cma(6, gap_cma_objective, 3.0, init, false, true);
    CMA<double,double> cma(5, gap_cma_objective, 3.0, init, false, true);
    double old = cma.mean_y;
    for (int i=0; i<20000; i++) {
        cma.step();
        if (i%20==0) {    
            if (abs(cma.mean_y-old)/old < 1e-4)
                break;
            old = cma.mean_y;
         cout << "n=" << globN <<  ", step " << i << ", " << cma  << endl;
        }
    }

    // format output
    cout << endl;
    cout << "OptimizationO" << globN << "= 5 parameters" << endl;
    cout << "cv2O" << globN << "=" << cma.mean_y << endl;
    cout << "cv2-1.5O" << globN << "=" << std::pow(2*globN+1,-1.5)  << endl;

//ascale, place, gap, pmax, istar, Ascale
    double ascale = abs(cma.mean[0]);
    double place = abs(cma.mean[1]);
    double gap = abs(cma.mean[2]);
    
    double pmax = abs(cma.mean[3]);
//    int istar = floor(abs(cma.mean[4]));
//    double Ascale = abs(cma.mean[5]);
    double Ascale = abs(cma.mean[4]);


    
    double pmin = 1.0/globN; 

    cout << "ascaleO" << globN << "=" << ascale << ", init=" << init[0] << endl;
    cout << "placeO" << globN << "=" << place << ", init=" << init[1] << endl;
    cout << "gapO" << globN << "=" << gap << ", init=" << init[2] << endl;
    cout << "pmaxO" << globN << "=" << pmax << ", init=" << init[3] << endl;
    cout << "istarO" << globN << "=" << istar <<  endl;
    cout << "AscaleO" << globN << "=" << Ascale << ", init=" << init[4] << endl;
    cout << "pminO" << globN << "= 1.0/" << globN << "= " << pmin << endl;

    cout << "{n,min,max,p,w,istar,a,A}={" << globN << ", " << pmin << ", " << pmax << ", " << place << ", " << 2*gap  << ", " << istar << ", " << ascale << ", " << Ascale << "}"  << endl;

    cout << "phiO" << globN << "={";
    
//    double pmax = 3.8 * globN; 
//    double Ascale = 4;
//    int istar =  globN-floor(std::sqrt(globN/1.0));
    double pw = 1.0/ascale;
    
    double lambda = istar * std::pow(pmin,pw) - std::pow(place-gap,pw) ;
    lambda = lambda / ( std::pow(pmin,pw) - std::pow(place-gap,pw) ) ; 
    
    double buff = (1.0 - istar ) * std::pow(pmin,pw) / ( std::pow(pmin,pw) - std::pow(place-gap,pw) ) ;
    double tau = pmin * std::pow(buff, - ascale) ; 
    
    vector<mpreal> phi;
    for (int j=1; j<=istar; j++)
        phi.push_back( tau * std::pow( j - lambda , ascale) );
 
 
    if(istar<globN-1){
    buff= std::pow(istar+1,Ascale)-std::pow(globN,Ascale);
    double Lambda = log((place+gap)/pmax)/buff;
    buff= std::pow(istar+1,Ascale)*log((place+gap)/pmax)/buff;
    double Theta = buff - log((place+gap));
    for (int j=istar+1; j<=globN; j++)
            phi.push_back( exp(Lambda * std::pow(j, Ascale) - Theta )  );
    }
    if(istar==globN-1){
    phi.push_back(pmax);
    }

    for (int i=0; i<globN; i++) {
        cout << phi[i];
        if (i==globN-1)
            cout << "}" << endl << endl;
        else
            cout << ", ";
    }
 }

void standard_cmaes() {

    double init[globN];
    for (int i=0; i<globN; i++)
        init[i] = i+1;
    
    int upoints=floor(std::sqrt(globN/1.0));
    for (int i=0; i<upoints; i++)
        init[globN-1-i] = globN * (4 - i/upoints ) ;    

    CMA<double,double> cma(globN, cma_objective, 3.0, init, false, true);
    double old = cma.mean_y;
    for (int i=0; i<20000; i++) {
        cma.step();        
        if (i%20==0) {
            if (abs(cma.mean_y-old)/old<1e-8)
                break;
            old = cma.mean_y;
        }
    }

    // format output
    cout << "cv2O" << globN << "=" << cma.mean_y << endl;
    cout << "phiO" << globN << "={";
    vector<double> phi;
    for (int i=0; i<globN; i++) {
        double val = cma.mean[i];
        if (val<0)
            val=globN;
        phi.push_back(val);
    }
    sort(phi.begin(), phi.end());
    for (int i=0; i<globN; i++) {
        cout << phi[i];
        if (i==globN-1)
            cout << "}" << endl << endl;
        else
            cout << ", ";
    }
    
    mpreal::set_default_prec(mpfr::digits2bits(precision));
    vector<mpreal> params(globN);
    for (int i=0; i<globN; i++)
        params[i] = mpreal(phi[i],precision);
        
    HyperPoly mypoly(2*globN+1);  
    compPolyCoeffs(params, mypoly);

}

ProgressFunc<CMAParameters<>,CMASolutions> select_time = [](const CMAParameters<> &cmaparams, const CMASolutions &cmasols)
{
  if (cmasols.niter() % 100 == 0)
    std::cerr << cmasols.elapsed_last_iter() << ": " << cmasols << std::endl;
  return 0;
};

void bipop_cmaes() {

    std::vector<double> init(globN);
    for (int i=0; i<globN; i++)
        init[i] = 3.8*globN*rand()/RAND_MAX;

    double pop_size_multiplier = 3.0;
    CMAParameters<> cmaparams(init,0.1,pop_size_multiplier*(4+3*log((double)(globN+1))));
    cmaparams.set_restarts(5);
    cmaparams.set_algo(BIPOP_CMAES);
//    CMASolutions cmasols = cmaes<>(bipop_cma_objective,cmaparams,select_time);
    CMASolutions cmasols = cmaes<>(bipop_cma_objective,cmaparams);
    std::cout << "* " << cmasols << std::endl;
    std::cout << "* optimization took " << cmasols.elapsed_time() / 1000.0 << " seconds\n";

    // format output
    cout << "cv2O" << globN << "=" << cmasols.best_candidate().get_fvalue() << endl;
    cout << "phiO" << globN << "={";
    vector<double> phi = cmasols.best_candidate().get_x();
    sort(phi.begin(), phi.end());
    for (int i=0; i<globN; i++) {
        cout << phi[i];
        if (i==globN-1)
            cout << "}" << endl;
        else
            cout << ", ";
    }
}



int main()
{

//    int points[6]={15,20,30,50,70,100};
//    std::cout.precision(6);

//    int points[6]={150,200,300,500,700,1000};
//    std::cout.precision(12);

    int points[4]={100,10000,7000,5000};
//    std::cout.precision(20);

    
    mpfact[0]= mpOne;
    mpreal mptemp = mpOne;
    for(int i=1; i<=2*Nmax+3; i++)
    {
 	mpfact[i]=mpfact[i-1]*mptemp;
    	mptemp += mpOne;
    }
  
//    cout << "factorials are computed!" << endl;

    srand((int)time(NULL));



     for (int N=0; N<4; N++) {
//     for (int N=1500; N<=2000; N+=500) {


//        precision = (50+ceil(1.487+N*0.647))*5;
//        mpreal::set_default_prec(mpfr::digits2bits(precision));


        if(Nmax<points[N]) {cout << "Nmax<N !!!" << endl; return 1;};
        globN = points[N];
//        if(Nmax<N) {cout << "Nmax<N !!!" << endl; return 1;};
//        globN = N;



//      Select appropriate optimization method: bipop is the slowest but gives the best results, gap_cmaes is the fast heuristic proposed in our paper
//        bipop_cmaes();
//        standard_cmaes();
        gap_cmaes();
    }

    return 0;
}
