(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[    145885,       3120]
NotebookOptionsPosition[    144988,       3088]
NotebookOutlinePosition[    145330,       3103]
CellTagsIndexPosition[    145287,       3100]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
     RowBox[{
     "Numerical", " ", "inverse", " ", "Laplace", " ", "transformation", " ", 
      "method"}], ";", "\[IndentingNewLine]", "PARAMETERS", ";", 
     "\[IndentingNewLine]", 
     RowBox[{"-", 
      RowBox[{"LTFunc", ":", " ", 
       RowBox[{
       "Laplace", " ", "domain", " ", "function", " ", "to", " ", "be", " ", 
        "inverse", " ", "transformed"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"-", 
      RowBox[{"t", ":", " ", 
       RowBox[{
       "Point", " ", "of", " ", "interest", " ", "for", " ", "time", " ", 
        "domain", " ", "function"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"-", 
      RowBox[{"order", ":", " ", 
       RowBox[{"Order", " ", "of", " ", "NILT", " ", 
        RowBox[{"(", 
         RowBox[{
         "number", " ", "of", " ", "evaluations", " ", "needed", " ", "for", 
          " ", "a", " ", "single", " ", "point"}], ")"}]}]}]}], ";", 
     "\[IndentingNewLine]", "\t", 
     RowBox[{
      RowBox[{"max", " ", "201", " ", "for", " ", "CME"}], "-", "R"}]}], ",", 
    " ", 
    RowBox[{"max", " ", "1000", " ", "for", " ", "CME"}], ",", " ", 
    RowBox[{
     RowBox[{"unlimited", " ", "for", " ", "other", " ", "methods"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"-", 
      RowBox[{"ILTmethod", ":", " ", 
       RowBox[{"ILT", " ", "method", " ", "to", " ", "be", " ", 
        RowBox[{"used", ".", " ", "Options"}]}]}]}], ";", 
     "\[IndentingNewLine]", "\t", 
     RowBox[{"\"\<Euler\>\"", " ", 
      RowBox[{"(", 
       RowBox[{"Euler", " ", "method"}], ")"}]}], ";", "\[IndentingNewLine]", 
     "\t", 
     RowBox[{"\"\<CME\>\"", " ", 
      RowBox[{"(", 
       RowBox[{
       "CME", " ", "method", " ", "with", " ", "complex", " ", 
        "eigenvalues"}], ")"}]}], ";", "\[IndentingNewLine]", "\t", 
     RowBox[{"\"\<GaverSt\>\"", " ", 
      RowBox[{"(", 
       RowBox[{"Gaver", "-", 
        RowBox[{"Stehfest", " ", "method"}]}], ")"}]}], ";", 
     "\[IndentingNewLine]", "\t", 
     RowBox[{"\"\<CMER\>\"", " ", 
      RowBox[{"(", 
       RowBox[{
       "CME", " ", "method", " ", "with", " ", "real", " ", "eigenvalues"}], 
       ")"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"-", 
      RowBox[{"prec", ":", " ", 
       RowBox[{
        RowBox[{
        "Required", " ", "precision", " ", "of", " ", "internal", " ", 
         "calculation", " ", "in", " ", "digits"}], " ", "-", " ", 
        RowBox[{
        "only", " ", "used", " ", "for", " ", "\"\<Euler\>\"", " ", "and", 
         " ", "\"\<GaverSt\>\"", " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"for", " ", "recommended", " ", "values"}], ",", " ", 
           RowBox[{"see", " ", "the", " ", "paper"}]}], ")"}]}]}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"OUTPUT", " ", "VALUE"}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"-", "Inverse"}], " ", "transformation", " ", "results", " ", 
      "in", " ", "point", " ", "t"}], ";", "\[IndentingNewLine]", 
     "REQUIREMENTS", ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"-", "For"}], " ", "\"\<CME\>\""}]}], ",", " ", 
    RowBox[{
     RowBox[{"the", " ", 
      RowBox[{"iltcme_ext", ".", "json"}], " ", "file", " ", "has", " ", "to",
       " ", "be", " ", "in", " ", "the", " ", "same", " ", "directory", " ", 
      "as", " ", 
      RowBox[{"this", " ", ".", "nb"}], " ", "file"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"-", "For"}], " ", "\"\<CMER\>\""}]}], ",", " ", 
    RowBox[{
     RowBox[{"the", " ", 
      RowBox[{"cmeretabeta", ".", "mx"}], " ", "file", " ", "has", " ", "to", 
      " ", "be", " ", "in", " ", "the", " ", "same", " ", "directory", " ", 
      "as", " ", 
      RowBox[{"this", " ", ".", "nb"}], " ", "file"}], ";"}]}], 
   "\[IndentingNewLine]", " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"NILT", "[", 
      RowBox[{"LTFunc_", ",", "t_", ",", "order_", ",", "ILTmethod_", ",", 
       RowBox[{"prec_:", "MachinePrecision"}]}], "]"}], ":=", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
        "\[Eta]", ",", "\[Beta]", ",", "M", ",", "l", ",", "f", ",", 
         "start"}], "}"}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"If", "[", " ", 
         RowBox[{
          RowBox[{"ILTmethod", "\[Equal]", "\"\<Euler\>\""}], ",", 
          RowBox[{"l", "=", 
           RowBox[{"Euler", "[", 
            RowBox[{"order", ",", "prec"}], "]"}]}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", " ", 
         RowBox[{
          RowBox[{"ILTmethod", "\[Equal]", "\"\<CME\>\""}], ",", 
          RowBox[{"l", "=", 
           RowBox[{"CME", "[", "order", "]"}]}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", " ", 
         RowBox[{
          RowBox[{"ILTmethod", "\[Equal]", "\"\<GaverSt\>\""}], ",", 
          RowBox[{"l", "=", 
           RowBox[{"GaverSt", "[", 
            RowBox[{"order", ",", "prec"}], "]"}]}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", " ", 
         RowBox[{
          RowBox[{"ILTmethod", "\[Equal]", "\"\<CMER\>\""}], ",", 
          RowBox[{"l", "=", 
           RowBox[{"CMER", "[", "order", "]"}]}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"\[Eta]", "=", 
         RowBox[{"l", "[", 
          RowBox[{"[", "1", "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"\[Beta]", "=", 
         RowBox[{"l", "[", 
          RowBox[{"[", "2", "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"M", "=", 
         RowBox[{
          RowBox[{"Length", "[", "\[Eta]", "]"}], "-", "1"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"\[Eta]", "[", 
            RowBox[{"[", "1", "]"}], "]"}], "\[Equal]", "0"}], ",", 
          RowBox[{"start", "=", "1"}], ",", 
          RowBox[{"start", "=", "0"}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"f", "=", 
         RowBox[{
          FractionBox["1", "t"], "  ", 
          RowBox[{"Sum", "[", 
           RowBox[{
            RowBox[{"Re", "[", 
             RowBox[{
              RowBox[{"\[Eta]", "[", 
               RowBox[{"[", 
                RowBox[{"k", "+", "1"}], "]"}], "]"}], " ", 
              RowBox[{"LTFunc", "[", 
               FractionBox[
                RowBox[{"\[Beta]", "[", 
                 RowBox[{"[", 
                  RowBox[{"k", "+", "1"}], "]"}], "]"}], "t"], " ", "]"}]}], 
             "]"}], ",", 
            RowBox[{"{", 
             RowBox[{"k", ",", "start", ",", "M"}], "}"}]}], "]"}]}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"Return", "[", "f", "]"}], ";"}]}], "\[IndentingNewLine]", 
      "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"Euler", "[", 
      RowBox[{"M_", ",", "prec_"}], "]"}], ":=", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"\[Xi]", ",", "\[Eta]", ",", "\[Beta]", ",", "k", ",", "n"}], 
        "}"}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"n", "=", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"Mod", "[", 
             RowBox[{"M", ",", "2"}], "]"}], "\[Equal]", "0"}], ",", 
           RowBox[{"M", "-", "1"}], ",", "M"}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"\[Beta]", "=", 
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"SetPrecision", "[", 
            RowBox[{
             RowBox[{
              FractionBox[
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"n", "-", "1"}], ")"}], 
                RowBox[{"Log", "[", "10", "]"}]}], "6"], "+", 
              RowBox[{"\[Pi]", " ", "\[ImaginaryI]", " ", 
               RowBox[{"(", 
                RowBox[{"k", "-", "1"}], ")"}]}]}], ",", "prec"}], "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"\[Xi]", "=", 
         RowBox[{"ConstantArray", "[", 
          RowBox[{
           RowBox[{"SetPrecision", "[", 
            RowBox[{"1", ",", "prec"}], "]"}], ",", "n"}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"\[Xi]", "[", 
          RowBox[{"[", "1", "]"}], "]"}], " ", "=", " ", 
         RowBox[{"SetPrecision", "[", 
          RowBox[{
           RowBox[{"1", "/", "2"}], ",", "prec"}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"\[Xi]", "[", 
          RowBox[{"[", "n", "]"}], "]"}], "=", 
         RowBox[{"SetPrecision", "[", 
          RowBox[{
           FractionBox["1", 
            SuperscriptBox["2", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{"n", "-", "1"}], ")"}], "/", "2"}]]], ",", "prec"}], 
          "]"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"For", "[", 
         RowBox[{
          RowBox[{"k", "=", "1"}], ",", 
          RowBox[{"k", "<", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"n", "-", "1"}], ")"}], "/", "2"}]}], ",", 
          RowBox[{"k", "++"}], ",", 
          RowBox[{
           RowBox[{"\[Xi]", "[", 
            RowBox[{"[", 
             RowBox[{"n", "-", "k"}], "]"}], "]"}], "=", 
           RowBox[{
            RowBox[{"\[Xi]", "[", 
             RowBox[{"[", 
              RowBox[{"n", "-", "k", "+", "1"}], "]"}], "]"}], "+", " ", 
            RowBox[{
             SuperscriptBox["2", 
              RowBox[{
               RowBox[{"-", 
                RowBox[{"(", 
                 RowBox[{"n", "-", "1"}], ")"}]}], "/", "2"}]], 
             RowBox[{"Binomial", "[", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"n", "-", "1"}], ")"}], "/", "2"}], ",", "k"}], 
              "]"}]}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"\[Eta]", "=", 
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"SetPrecision", "[", 
            RowBox[{
             RowBox[{
              SuperscriptBox["10", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"n", "-", "1"}], ")"}], "/", "6"}]], 
              SuperscriptBox[
               RowBox[{"(", 
                RowBox[{"-", "1"}], ")"}], 
               RowBox[{"k", "-", "1"}]], 
              RowBox[{"\[Xi]", "[", 
               RowBox[{"[", "k", "]"}], "]"}]}], ",", "prec"}], "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"\[Eta]", "=", 
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"SetPrecision", "[", 
             RowBox[{
              RowBox[{
               SuperscriptBox["10", 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{"n", "-", "1"}], ")"}], "/", "6"}]], 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"-", "1"}], ")"}], 
                RowBox[{"k", "-", "1"}]]}], ",", "prec"}], "]"}], "*", 
            RowBox[{"\[Xi]", "[", 
             RowBox[{"[", "k", "]"}], "]"}]}], ",", 
           RowBox[{"{", 
            RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"Return", "[", 
         RowBox[{"{", 
          RowBox[{"\[Eta]", ",", "\[Beta]"}], "}"}], "]"}], ";"}]}], 
      "\[IndentingNewLine]", "\[IndentingNewLine]", "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"GaverSt", "[", 
      RowBox[{"M_", ",", "prec_"}], "]"}], ":=", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"\[Eta]", ",", "\[Beta]", ",", "n"}], "}"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"n", "=", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"Mod", "[", 
             RowBox[{"M", ",", "2"}], "]"}], "\[Equal]", "1"}], ",", 
           RowBox[{"M", "-", "1"}], ",", "M"}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"\[Beta]", "=", 
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"k", " ", 
            RowBox[{"SetPrecision", "[", 
             RowBox[{
              RowBox[{"Log", "[", "2", "]"}], ",", "prec"}], "]"}]}], ",", 
           RowBox[{"{", 
            RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"\[Eta]", "=", 
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"SetPrecision", "[", 
             RowBox[{
              RowBox[{"Log", "[", "2", "]"}], ",", "prec"}], "]"}], 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"-", "1"}], ")"}], 
             RowBox[{
              RowBox[{"n", "/", "2"}], "+", "k"}]], 
            RowBox[{"Sum", "[", 
             RowBox[{
              RowBox[{
               FractionBox[
                SuperscriptBox["j", 
                 RowBox[{
                  RowBox[{"n", "/", "2"}], "+", "1"}]], 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{"n", "/", "2"}], ")"}], "!"}]], 
               RowBox[{"Binomial", "[", 
                RowBox[{
                 RowBox[{"n", "/", "2"}], ",", "j"}], "]"}], 
               RowBox[{"Binomial", "[", 
                RowBox[{
                 RowBox[{"2", "j"}], ",", "j"}], "]"}], 
               RowBox[{"Binomial", "[", 
                RowBox[{"j", ",", 
                 RowBox[{"k", "-", "j"}]}], "]"}]}], ",", 
              RowBox[{"{", 
               RowBox[{"j", ",", 
                RowBox[{"Floor", "[", 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{"k", "+", "1"}], ")"}], "/", "2"}], "]"}], ",", 
                RowBox[{"Min", "[", 
                 RowBox[{"k", ",", 
                  RowBox[{"n", "/", "2"}]}], "]"}]}], "}"}]}], "]"}]}], "  ", 
           ",", 
           RowBox[{"{", 
            RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"Return", "[", 
         RowBox[{"{", 
          RowBox[{"\[Eta]", ",", "\[Beta]"}], "}"}], "]"}], ";"}]}], 
      "\[IndentingNewLine]", "\[IndentingNewLine]", "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"CME", "[", "M_", "]"}], ":=", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
        "params", ",", "\[Eta]re", ",", "\[Eta]im", ",", "\[Beta]re", ",", 
         "\[Beta]im", ",", " ", "\[Eta]", ",", "\[Beta]", ",", "prec", ",", 
         "i"}], "}"}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"prec", "=", 
         RowBox[{
          RowBox[{"0.65", "M"}], "+", "20"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"Length", "[", 
             RowBox[{"Position", "[", 
              RowBox[{
               RowBox[{"Names", "[", "\"\<Global`*\>\"", "]"}], ",", 
               "\"\<cmeParams2\>\""}], "]"}], "]"}], "\[Equal]", "0"}], " ", "||",
            " ", 
           RowBox[{
            RowBox[{"Length", "[", "cmeParams2", "]"}], "\[Equal]", "0"}]}], 
          ",", "\[IndentingNewLine]", 
          RowBox[{"(*", 
           RowBox[{
            RowBox[{"Print", "[", "\"\<CME parameters loaded\>\"", "]"}], 
            ";"}], "*)"}], "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"cmeParams2", " ", "=", " ", 
            RowBox[{"Import", "[", 
             RowBox[{"FileNameJoin", "[", 
              RowBox[{"{", 
               RowBox[{
                RowBox[{"NotebookDirectory", "[", "]"}], ",", 
                "\"\<iltcme_ext.json\>\""}], "}"}], "]"}], "]"}]}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"nParams", "=", 
            RowBox[{"\"\<n\>\"", "/.", "cmeParams2"}]}], ";"}]}], 
         "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"For", "[", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", 
          RowBox[{"i", "\[LessEqual]", 
           RowBox[{"Length", "[", "nParams", "]"}]}], ",", 
          RowBox[{"i", "++"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{
              RowBox[{"nParams", "[", 
               RowBox[{"[", "i", "]"}], "]"}], "\[GreaterEqual]", "M"}], ",", 
             "\[IndentingNewLine]", "\t", 
             RowBox[{
              RowBox[{"params", "=", 
               RowBox[{"cmeParams2", "[", 
                RowBox[{"[", "i", "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
              "\t", 
              RowBox[{"\[Eta]re", "=", 
               RowBox[{"\"\<eta_re\>\"", "/.", "params"}]}], ";", 
              "\[IndentingNewLine]", "\t", 
              RowBox[{"\[Eta]im", "=", 
               RowBox[{"\"\<eta_im\>\"", "/.", "params"}]}], ";", 
              "\[IndentingNewLine]", "\t", 
              RowBox[{"\[Beta]re", "=", 
               RowBox[{"\"\<beta_re\>\"", "/.", "params"}]}], ";", 
              "\[IndentingNewLine]", "\t", 
              RowBox[{"\[Beta]im", "=", 
               RowBox[{"\"\<beta_im\>\"", "/.", "params"}]}], ";", 
              "\[IndentingNewLine]", "\t", 
              RowBox[{"\[Eta]", "=", 
               RowBox[{"SetPrecision", "[", 
                RowBox[{
                 RowBox[{
                  RowBox[{"{", 
                   RowBox[{"\[Eta]re", "[", 
                    RowBox[{"[", "1", "]"}], "]"}], "}"}], "~", "Join", "~", 
                  RowBox[{"Flatten", "[", 
                   RowBox[{
                    RowBox[{"Table", "[", 
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"\[Eta]re", "[", 
                    RowBox[{"[", "i", "]"}], "]"}], "+", 
                    RowBox[{"\[ImaginaryI]", "*", 
                    RowBox[{"\[Eta]im", "[", 
                    RowBox[{"[", "i", "]"}], "]"}]}]}], ",", 
                    RowBox[{
                    RowBox[{"\[Eta]re", "[", 
                    RowBox[{"[", "i", "]"}], "]"}], "-", 
                    RowBox[{"\[ImaginaryI]", "*", 
                    RowBox[{"\[Eta]im", "[", 
                    RowBox[{"[", "i", "]"}], "]"}]}]}]}], "}"}], ",", 
                    RowBox[{"{", 
                    RowBox[{"i", ",", "2", ",", 
                    RowBox[{"M", "+", "1"}]}], "}"}]}], "]"}], "/", "2"}], 
                   "]"}]}], ",", "prec"}], "]"}]}], ";", 
              "\[IndentingNewLine]", "\t", 
              RowBox[{"\[Beta]", "=", 
               RowBox[{"SetPrecision", "[", 
                RowBox[{
                 RowBox[{
                  RowBox[{"{", 
                   RowBox[{"\[Beta]re", "[", 
                    RowBox[{"[", "1", "]"}], "]"}], "}"}], "~", "Join", "~", 
                  RowBox[{"Flatten", "[", 
                   RowBox[{"Table", "[", 
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"\[Beta]re", "[", 
                    RowBox[{"[", "i", "]"}], "]"}], "+", 
                    RowBox[{"\[ImaginaryI]", "*", 
                    RowBox[{"\[Beta]im", "[", 
                    RowBox[{"[", "i", "]"}], "]"}]}]}], ",", 
                    RowBox[{
                    RowBox[{"\[Beta]re", "[", 
                    RowBox[{"[", "i", "]"}], "]"}], "-", 
                    RowBox[{"\[ImaginaryI]", "*", 
                    RowBox[{"\[Beta]im", "[", 
                    RowBox[{"[", "i", "]"}], "]"}]}]}]}], "}"}], ",", 
                    RowBox[{"{", 
                    RowBox[{"i", ",", "2", ",", 
                    RowBox[{"M", "+", "1"}]}], "}"}]}], "]"}], "]"}]}], ",", 
                 "prec"}], "]"}]}], ";", "\[IndentingNewLine]", "\t", 
              RowBox[{"Return", "[", 
               RowBox[{"{", 
                RowBox[{"\[Eta]", ",", "\[Beta]"}], "}"}], "]"}], ";"}]}], 
            "\[IndentingNewLine]", "]"}], ";"}]}], "\[IndentingNewLine]", 
         "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{
        "Print", "[", "\"\<Error! Requested order not available!\>\"", "]"}], 
        ";", "\[IndentingNewLine]", 
        RowBox[{"Return", "[", 
         RowBox[{"{", "}"}], "]"}], ";"}]}], "\[IndentingNewLine]", "]"}]}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"CMER", "[", "M_", "]"}], ":=", "\[IndentingNewLine]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "n", "}"}], ",", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"Length", "[", 
            RowBox[{"Position", "[", 
             RowBox[{
              RowBox[{"Names", "[", "\"\<Global`*\>\"", "]"}], ",", 
              "\"\<cmeretabeta\>\""}], "]"}], "]"}], "\[Equal]", "0"}], " ", "||",
           " ", 
          RowBox[{
           RowBox[{"Length", "[", "cmeretabeta", "]"}], "\[Equal]", "0"}]}], 
         ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"cmeretabeta", " ", "=", " ", 
           RowBox[{"Import", "[", 
            RowBox[{"FileNameJoin", "[", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"NotebookDirectory", "[", "]"}], ",", 
               "\"\<cmeretabeta.mx\>\""}], "}"}], "]"}], "]"}]}], ";"}]}], 
        "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"n", "=", 
        RowBox[{"Ceiling", "[", 
         RowBox[{"M", "/", "2"}], "]"}]}], ";", "\[IndentingNewLine]", 
       RowBox[{"Return", "[", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"SetPrecision", "[", 
           RowBox[{
            RowBox[{"cmeretabeta", "[", 
             RowBox[{"[", 
              RowBox[{"n", ",", "1"}], "]"}], "]"}], ",", 
            RowBox[{"Precision", "[", 
             RowBox[{"cmeretabeta", "[", 
              RowBox[{"[", 
               RowBox[{"n", ",", "1"}], "]"}], "]"}], "]"}]}], "]"}], ",", 
          RowBox[{"SetPrecision", "[", 
           RowBox[{
            RowBox[{"cmeretabeta", "[", 
             RowBox[{"[", 
              RowBox[{"n", ",", "2"}], "]"}], "]"}], ",", 
            RowBox[{"Precision", "[", 
             RowBox[{"cmeretabeta", "[", 
              RowBox[{"[", 
               RowBox[{"n", ",", "1"}], "]"}], "]"}], "]"}]}], "]"}]}], "}"}],
         "]"}], ";"}]}], "\[IndentingNewLine]", "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.808565090849353*^9, 3.808565092099605*^9}, {
  3.808574790034979*^9, 3.8085752846935954`*^9}, {3.8085753152244864`*^9, 
  3.808575415535327*^9}, {3.808575472004349*^9, 3.8085755703992863`*^9}, {
  3.8085756730949783`*^9, 3.8085756796075706`*^9}, {3.8085757873920546`*^9, 
  3.808575852487034*^9}, {3.8085758868322196`*^9, 3.808575891647346*^9}, {
  3.8085759589084897`*^9, 3.808575981868109*^9}, {3.808576166611802*^9, 
  3.8085762561007347`*^9}, {3.808576286948269*^9, 3.8085762996303663`*^9}, {
  3.808576363938629*^9, 3.808576366339244*^9}, {3.808648947789565*^9, 
  3.808648987166338*^9}}],

Cell[BoxData[""], "Input",
 CellChangeTimes->{{3.808565085370582*^9, 3.808565149015293*^9}, {
   3.8085652947479725`*^9, 3.8085652949344916`*^9}, {3.8085656938441224`*^9, 
   3.808565773251266*^9}, {3.808568838386233*^9, 3.808568990601323*^9}, {
   3.808569149414566*^9, 3.8085692070332775`*^9}, {3.8085692471480107`*^9, 
   3.8085693427165604`*^9}, {3.808575605240405*^9, 3.8085756580193167`*^9}, 
   3.8085759796949186`*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", "EXAMPLES", " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"Fexp", "[", "s_", "]"}], ":=", 
     FractionBox["1", 
      RowBox[{"1", "+", "s"}]]}], ";"}], " ", 
   RowBox[{"(*", " ", 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"-", "t"}]], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"Texp", "[", "t_", "]"}], ":=", 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"-", "t"}]]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"Fstep", "[", "s_", "]"}], ":=", 
     RowBox[{
      FractionBox["1", "s"], " ", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"-", " ", "s", " "}]]}]}], ";", " ", 
    RowBox[{"(*", " ", 
     RowBox[{"unit", " ", "step", " ", "function"}], " ", "*)"}], 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Tstep", "[", "t_", "]"}], ":=", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"t", ">", "1"}], ",", "1", ",", "0"}], "]"}]}], 
    ";"}]}]}]], "Input",
 CellChangeTimes->{{3.8085754448219895`*^9, 3.8085754603135834`*^9}, {
  3.808640208456489*^9, 3.8086402144165554`*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"order", "=", "50"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"precision", "=", "70"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"min", "=", 
   RowBox[{"1", "/", "100"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"max", "=", "3"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"step", "=", 
   RowBox[{"1", "/", "100"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"texp", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"Texp", "[", "t", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expgaver", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"NILT", "[", 
        RowBox[{
        "Fexp", ",", "t", ",", "order", ",", "\"\<GaverSt\>\"", ",", 
         "precision"}], " ", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expeuler", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"NILT", "[", 
        RowBox[{
        "Fexp", ",", "t", ",", "order", ",", "\"\<Euler\>\"", ",", 
         "precision"}], " ", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expcme", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"NILT", "[", 
        RowBox[{
        "Fexp", ",", "t", ",", "order", ",", "\"\<CME\>\"", ",", 
         "precision"}], " ", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"expcmer", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"NILT", "[", 
        RowBox[{
        "Fexp", ",", "t", ",", "order", ",", "\"\<CMER\>\"", ",", 
         "precision"}], " ", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"tstep", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"Tstep", "[", "t", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"stepgaver", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"NILT", "[", 
        RowBox[{
        "Fstep", ",", "t", ",", "order", ",", "\"\<GaverSt\>\"", ",", 
         "precision"}], " ", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"stepeuler", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"NILT", "[", 
        RowBox[{
        "Fstep", ",", "t", ",", "order", ",", "\"\<Euler\>\"", ",", 
         "precision"}], " ", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"stepcme", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"NILT", "[", 
        RowBox[{
        "Fstep", ",", "t", ",", "order", ",", "\"\<CME\>\"", ",", 
         "precision"}], " ", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"stepcmer", "=", 
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"t", ",", 
        RowBox[{"NILT", "[", 
         RowBox[{
         "Fstep", ",", "t", ",", "order", ",", "\"\<CMER\>\"", ",", 
          "precision"}], " ", "]"}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"t", ",", "min", ",", "max", ",", "step"}], "}"}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.808565315915402*^9, 3.808565317667681*^9}, {
  3.808565351070716*^9, 3.80856543359371*^9}, {3.8085655106292*^9, 
  3.8085655455642257`*^9}, {3.808569008300423*^9, 3.808569010233596*^9}, {
  3.8085759999807277`*^9, 3.808576005929816*^9}, {3.8086402266757836`*^9, 
  3.8086405228870397`*^9}, {3.808640573511776*^9, 3.8086406432444296`*^9}, {
  3.8086407061567717`*^9, 3.808640761278413*^9}, {3.808648968776496*^9, 
  3.808649002806489*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "texp", ",", "expgaver", ",", "expeuler", ",", "expcme", ",", "expcmer"}],
     "}"}], ",", 
   RowBox[{"PlotStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
     "Black", ",", " ", "Dashed", ",", "Green", ",", "Red", ",", "Blue"}], 
     "}"}]}], " ", ",", 
   RowBox[{"PlotLegends", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
     "\"\<f(t)\>\"", ",", " ", "\"\<Gaver\>\"", ",", "\"\<Euler\>\"", ",", 
      "\"\<CME\>\"", ",", "\"\<CMER\>\""}], "}"}]}], ",", 
   RowBox[{"ImageSize", "\[Rule]", "700"}]}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "tstep", ",", "stepgaver", ",", "stepeuler", ",", "stepcme", ",", 
     "stepcmer"}], "}"}], ",", 
   RowBox[{"PlotStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
     "Black", ",", " ", "Dashed", ",", "Green", ",", "Red", ",", "Blue"}], 
     "}"}]}], " ", ",", 
   RowBox[{"PlotLegends", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
     "\"\<f(t)\>\"", ",", " ", "\"\<Gaver\>\"", ",", "\"\<Euler\>\"", ",", 
      "\"\<CME\>\"", ",", "\"\<CMER\>\""}], "}"}]}], ",", 
   RowBox[{"ImageSize", "\[Rule]", "700"}]}], "]"}]}], "Input",
 CellChangeTimes->{{3.808640539334071*^9, 3.808640566597261*^9}, {
   3.8086406470662365`*^9, 3.808640692894704*^9}, {3.8086407765974603`*^9, 
   3.8086408917886314`*^9}, {3.8086409333096333`*^9, 3.808640946225112*^9}, {
   3.8086409998133698`*^9, 3.8086410129614925`*^9}, 3.808641057647025*^9, {
   3.808649009659165*^9, 3.8086490195706673`*^9}}],

Cell[BoxData[
 TemplateBox[{GraphicsBox[{{}, {{{}, {}, {
        Hue[0.67, 0.6, 0.6], 
        Directive[
         PointSize[0.008333333333333333], 
         AbsoluteThickness[1.6], 
         GrayLevel[0]], 
        LineBox[CompressedData["
1:eJw10nk81PkfB/CZMQdKW5JKx1Ibko1SybLen6La2nRIdOhw1W8ViXTYslvL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         "]]}, {
        Hue[0.9060679774997897, 0.6, 0.6], 
        Directive[
         PointSize[0.008333333333333333], 
         RGBColor[0.880722, 0.611041, 0.142051], 
         AbsoluteThickness[1.6], 
         Dashing[{Small, Small}]], 
        LineBox[CompressedData["
1:eJw10nk81PkfB/CZMQdKW5JKx1Ibko1SybLen6La2nRIdOhw1W8ViXTYslvL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         "]]}, {
        Hue[0.1421359549995791, 0.6, 0.6], 
        Directive[
         PointSize[0.008333333333333333], 
         AbsoluteThickness[1.6], 
         RGBColor[0, 1, 0]], 
        LineBox[CompressedData["
1:eJw10nlczFsfB/DZpyJXaJGLsiTpKkIifQ+SrVCoLqFUPDdFyZLQtdSVnRuR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         "]]}, {
        Hue[0.37820393249936934`, 0.6, 0.6], 
        Directive[
         PointSize[0.008333333333333333], 
         AbsoluteThickness[1.6], 
         RGBColor[1, 0, 0]], 
        LineBox[CompressedData["
1:eJw10ndclFcaBeBhZphRRIMGFBGJmAiiRAULouB7UbOga6wIGlGJiBrFghql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         "]]}, {
        Hue[0.6142719099991583, 0.6, 0.6], 
        Directive[
         PointSize[0.008333333333333333], 
         AbsoluteThickness[1.6], 
         RGBColor[0, 0, 1]], 
        LineBox[CompressedData["
1:eJw11nlczekeB/CztRCSmpgGERLJEmEo34crk8jSoqnBjSYNsoRrKXtMTJRS
gxjLFJM2LVLaT/tK+669Tvu+arvqfu7zz3lV5/c8z+fz/vY6Z/7hU/oWPA6H
s4zL4Yy/2sr5/1Rlc4vau5YfYkHt9L+fXSmzZUDF7lA7hc293bw37DntmDXo
7irz/797kvz9Vx5n4trI/cn4ekt3LOPW/HCuDe/3p9kFttr3lNuoZe/4bwJp
746VOfHFrXj+Pb3NPV8Wc7+VJhd6nNaaHEoGTlOPX93Siv0+0qbdGXdH+lto
yX6nuF6VSJIzerR9o38L9o+mWg39WA2LFvL89nShRywJKnIkm2e34DwhNUTm
bTHJb6bVEyuervJbXK46NuP8BMqceUnJRKeZ4ntVvp2QSCcVnimKeM24TzI5
u6hEKUc3kf7EgSmk9eMTdXmbJtwvlcxr+ZHBa5uoymZ8wzRycp7qPtLdiPtm
UJxTuUxXYCNZj1/vdCZtTpE/4XS6Eff/RO2L4gZzVzQSZ2JlUZ+C3uTo9gbk
ySKLrZd6dwQ0kM7EAdlkpaovf9m6AfmyibdQKWPn6gYaP22/Uw7Z61jaRPaK
kDeX1rqk3EkNE9FEnMJcGp0qZX7msgj588hJTeJSCBOR4vh15ubTljXPNzmL
idBHPrnHvFMVpNfTkYlVQHLHpSOnONejn0IadLQ59da4nsZ3k/MvpIHvHVOL
5tajryJafTty75P6Ouob3663iOyq//Ww0b8O/RWTj2DhnJQLdTRRj1YJaQjl
bq7eXIc+Syh8x+Wfl0rV0cT1bpfSTm+LDT75tei3jOK2XEh8/7KWMjPGVxkF
XG9017aqRd9fKOjhlX/vX19LE3HlyqlC+1xEv6AW/ZeTctMtU/mcGprYbn8F
scN5zR9e1MCjkqZrtswvOlGDPJWkWh8YdFGzBj6VpDt0wPiRVA3yVdIxz7MZ
6qXV8KoiM5mw7O0+1chbRb4SS5eW2lbDr4o22w8+rdtZjfzVNOv2tQ7LudXw
rKY9ERo9Zh1V6KOa/nL0P/85rgq+NVQgeV83yK0K/dSQZcNlbcmjVfCuoRMJ
S1mWZhX6qiV9bU35STJV8K+ll9Y6bwLrKtFfLYmmZWWnhFdiHurIVn7tFb0H
leizjoq0wpw2HanEfNSRlYtlt7tmJfqtpwCV+W6HZCsxL/V0JXJqqlJ+Bfqu
pwcHOv7Rv1WB+RGRxTr5wD1rKtC/iH7ZOTlPoa4c8ySiH0K9xN79WQ6PBros
ms6ktpfDo4E28h/fVBv+Ao8Gmn8sIls24As8GmjzAdl1kb9+gUcjtaZoCxcq
fIFHI01aNWi3I6sMHo1kGZP8aLl9GTyaKP2sukLGpjJ4NNHm/7iqKfWXwqOJ
zt6SH1weUAqPZnrt+LdXx9FSeDTTsFmhlcXCUng0U4rtsI1DRQk8WuicWN+U
Q09L4NFCqWXrD9Ual8CjhdanRYfLflcCj1Yq+PzVrDWnGB6tFPYo88Up52J4
tJIFxyb58Z5ieLTR5OQnyyynF8OjjZR3z1ApziqCRxsdtJ80s9W5CB7ttLK5
xPSVQRE82mntqykmPd8VwaOdAkK22tYWFcKjg6ZZb1px+lkh8nTQkrSYejez
Qvh0UFu340zDRYXI10GLxX/b+LapAF4dpCzs+vwwoAB5O4hvFbZvxoUC+HWQ
Wub3Rgs2FSB/J0lpaRsmihXAs5PMcpZV9WXmo49O0vC6fO/Dn/nw7aRpWs5N
PLN89NNJss0nHpYsyYd3J/EvBStp9uShry56Hv/nOuWYPPh3kfG1wnOP/shD
f130VSb0uuO+PMxDF8251S6YvCAPfXbR9D+8/SU7cjEfXSTteXH53ahc9NtN
vxvprP/DIRfz0k1yZw/YTTbNRd/dtMpeJ01ySS7mp5u+1ytPsxvIQf/dVLZ5
xTablBzMU/e3z9NrnObHOfDoIf9zWsK8oznw6CG32CH9tRtz4NFDhtMt7aZN
zYFHDy2bRwv2V2TDo4fMH+8SnxuUDY8eqlI35xvczoZHD2n2avcM/pwNj14q
uxocIqOWDY9emn/ccPETbjY8eulfySGqTgVZ8OilxKebvbp8suDRSxvvrrsd
dyMLHr206eJrvxHjLHj0kfF+M2mP5Vnw6COvSSPugWJZ8Oijgxtc7PeEf4ZH
H+W22Y9JHfsMjz4ydVnh9FDhMzz6iGexSK0u/RM8+smjnvJHr3yCRz/9/svN
O+UrP8Gjn6ack9a7WZsJj34ql5+9vOFxJjz66aBe9cZpuzLh0U8cfVeHLn4m
PAZoYezudQ/DM+AxQCp10w+2WWfAY4BM1N8s4C3NgMcAWX72CMioTofHAE3R
M1bf+SwdHgPkdiGi5Oq+dHgMULuOVeVhmXR4DFL/oPvVvow0eAySjzqvd93d
NHgMUu/PJfaLtqXBY5CiRjKvRfHT4DFIA5ZtS0aFqfAYJFMtr8K666nw+Eqn
dMLKz7BUeHylmF/mPfHgpMLjK0XunW1xXpgCj6+kvGTm68abKfD4Sp5aL2OG
tqbA4ytte2am6CORAo8henZ2+uq+tGR4DNHqF1YnSh2T4TFEsruWaZsYJMNj
iKQT+tacnJUMjyF6rGESMas8CR5DlBkVaWrsmQSPYbI7uf+JyvEkeAyTlM1G
roN6EjyG6dfzP8y78jURHsPk+fGM/ci374X/8ximvZ8PZ0y+lwiPYToi/s7e
wygRHsOUMTPYKEkxER4jNFxrUmDdlACPEXq6S87PMyQBHiMUq5R1/cCNBHiM
ULBS+5C7XgI8RkikuCnjF4UEeIxQ/YqWv5+L4uExSpL9N9YcDomHx+i3/4/i
9a/t4uExSq/NZY8f0Y+Hxyj19Lrc9ZwfD49vz+t2Gh3sjIPHKBVu6Hd1EcbB
Y4y2x7tOYy5x8Bij3Kqi+xbmcfAYI3HnxkquRhw8xkiUN1QkKxEHjzFKL9uh
614shMcY3f3xuymuvkJ4cFj+YoeBsWtC5OEw27asljIDIXw47MCerpQlKkLk
47AumbJTdcOx8OKwWS6GydNyYpGXwyad7wn1/CcWfhym+Zui2psrscjPYZJN
w3IzDGPhyWEZ66zNREtj0QeH3SR5/iJeLHw5rFz/aP3D4Bj0w2ExdZcMvM1j
4M1h+ho3tjrIxaAvLvsnuiBkTVI0/Lns1zmDXp4Xo9EflzlkXl9YrBqNeeAy
K2nZNVkVUeiTy6xbDRvsXaMwH1x2xrx83+j2KPTLZSv+UXNfxYnCvHDZC2f3
bMXQSPTNZQkquvOTT0Zifrjs9XaloEWLI9E/lymdiQr8sTIC88Rlu70P6nPd
I+DBY7z7dWk2hhHw4LEtp+L2eUpHwIPH9N40aVxOD4cHj31pzvDl3AmHB4+Z
V2oLVmqHw4PHWrxd70ryw+HBYwc+7Ha8I/wIDx4TOpj85n/9Izx4zCg13/oi
+wgPHpu2du6kOs5HePCYfmfogy5hGDx4zDVa59QLuzB48NjvVvekm7TD4MFn
Psa6lVmSYfDgs16JHYb6GaHw4DPlRr1qqweh8OCzqDljAkWjUHjwmbxV6wZL
hVB48JmL9nXxrZUf4MFnJ0WtrcFvPsCDz6TcVtmGnPgADz4728zV/EnjAzz4
LFkvWWg5EgIPPjt8zOXVd0kh8OCz34pmnNvtFAIPAbuXeT9f1iQEHgKW1Jxn
cWhBCDwELFxZK2Vt23t4CNjLWDcPl4/v4SFgpgcV/z57+z08BMxN3c66aO97
eAjYK0qLi5n7Hh4CVvzOUFelJRgeAma4NTtmengwPAQsoHhN+YU7wfAQMD9e
kPl+42B4CNh/wgJkopSD4SFgW+cl+T7uC4KHGNt1xHe4NSkIHmLMzW9DRsSj
IHiIsXtyBxo5R4PgIca2fZqyKHJDEDzEGG+PjkHLlCB4iLFiF6kdDyoC4SHG
zI/q1L8JCoSHGJNvLOeu+D0QHmJsmcjtymLTQHiIsdY7sppOywPhIcbkRPlK
FvxAeHx7f4KqgldRADzE2XsjtanG/gHwEGd5iiGNp28FwEOcWdf8/LLHNAAe
4sz2+eGFVasC4CHONJ+nHteYFAAPcTa0zbi8UvgOHuLszcp7Bkpn3sFDnKX5
l6urLXgHD3HGmzTvyliePzzE2auW9esf2fvD49t+h2af6trgDw9xplsTvkq6
3Q8e4kzo2uva7OEHDwnm8Jdz0G0TP3hIMINXqm9Lpf3gIcEkTVWfNSf5wkOC
3dkbGhx81RceEqw83Gu2+jpfeEiw+GPXJC06fOAhwWa0G/npevvAQ4KVJORr
V/3qAw8JluarKrN4ng88JNjTOcP6CmXe8JBgK/OMDT4+9oaHBFO+EXCQt88b
HpLs2sPMxG5Zb/ovkEQBlA==
         "]]}}}, {}, {}, {{}, {}}}, {
    DisplayFunction -> Identity, PlotRangePadding -> {{
        Scaled[0.02], 
        Scaled[0.02]}, {
        Scaled[0.02], 
        Scaled[0.05]}}, AxesOrigin -> {0, 0}, 
     PlotRange -> {{0, 3.}, {0, 0.9905092142213031}}, PlotRangeClipping -> 
     True, ImagePadding -> All, DisplayFunction -> Identity, AspectRatio -> 
     NCache[GoldenRatio^(-1), 0.6180339887498948], Axes -> {True, True}, 
     AxesLabel -> {None, None}, AxesOrigin -> {0, 0}, DisplayFunction :> 
     Identity, Frame -> {{False, False}, {False, False}}, 
     FrameLabel -> {{None, None}, {None, None}}, 
     FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
     GridLines -> {None, None}, GridLinesStyle -> Directive[
       GrayLevel[0.5, 0.4]], ImageSize -> 700, 
     Method -> {"CoordinatesToolOptions" -> {"DisplayFunction" -> ({
           (Part[{{Identity, Identity}, {Identity, Identity}}, 1, 2][#]& )[
            Part[#, 1]], 
           (Part[{{Identity, Identity}, {Identity, Identity}}, 2, 2][#]& )[
            Part[#, 2]]}& ), "CopiedValueFunction" -> ({
           (Part[{{Identity, Identity}, {Identity, Identity}}, 1, 2][#]& )[
            Part[#, 1]], 
           (Part[{{Identity, Identity}, {Identity, Identity}}, 2, 2][#]& )[
            Part[#, 2]]}& )}}, 
     PlotRange -> {{0, 3.}, {0, 0.9905092142213031}}, PlotRangeClipping -> 
     True, PlotRangePadding -> {{
        Scaled[0.02], 
        Scaled[0.02]}, {
        Scaled[0.02], 
        Scaled[0.05]}}, Ticks -> {Automatic, Automatic}}],FormBox[
    FormBox[
     TemplateBox[{
      "\"f(t)\"", "\"Gaver\"", "\"Euler\"", "\"CME\"", "\"CMER\""}, 
      "LineLegend", DisplayFunction -> (FormBox[
        StyleBox[
         StyleBox[
          PaneBox[
           TagBox[
            GridBox[{{
               TagBox[
                GridBox[{{
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    AbsoluteThickness[1.6], 
                    GrayLevel[0]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    AbsoluteThickness[1.6], 
                    GrayLevel[0]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    AbsoluteThickness[1.6], 
                    Dashing[{Small, Small}]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    AbsoluteThickness[1.6], 
                    Dashing[{Small, Small}]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #2}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0, 1, 0]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0, 1, 0]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #3}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    AbsoluteThickness[1.6], 
                    RGBColor[1, 0, 0]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    AbsoluteThickness[1.6], 
                    RGBColor[1, 0, 0]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #4}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0, 0, 1]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.075], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0, 0, 1]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #5}}, 
                 GridBoxAlignment -> {
                  "Columns" -> {Center, Left}, "Rows" -> {{Baseline}}}, 
                 AutoDelete -> False, 
                 GridBoxDividers -> {
                  "Columns" -> {{False}}, "Rows" -> {{False}}}, 
                 GridBoxItemSize -> {"Columns" -> {{All}}, "Rows" -> {{All}}},
                  GridBoxSpacings -> {
                  "Columns" -> {{0.5}}, "Rows" -> {{0.8}}}], "Grid"]}}, 
             GridBoxAlignment -> {"Columns" -> {{Left}}, "Rows" -> {{Top}}}, 
             AutoDelete -> False, 
             GridBoxItemSize -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
             GridBoxSpacings -> {"Columns" -> {{1}}, "Rows" -> {{0}}}], 
            "Grid"], Alignment -> Left, AppearanceElements -> None, 
           ImageMargins -> {{5, 5}, {5, 5}}, ImageSizeAction -> 
           "ResizeToFit"], LineIndent -> 0, StripOnInput -> False], {
         FontFamily -> "Arial"}, Background -> Automatic, StripOnInput -> 
         False], TraditionalForm]& ), 
      InterpretationFunction :> (RowBox[{"LineLegend", "[", 
         RowBox[{
           RowBox[{"{", 
             RowBox[{
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.008333333333333333`", "]"}], 
                   ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    GrayLevel[0], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> GrayLevel[0.], FrameTicks -> None, 
                    PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], "GrayLevel[0]"], 
                    Appearance -> None, BaseStyle -> {}, BaselinePosition -> 
                    Baseline, DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    GrayLevel[0]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["GrayLevelColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    GrayLevel[0], Editable -> False, Selectable -> False]}], 
                 "]"}], ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.008333333333333333`", "]"}], 
                   ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> 
                    RGBColor[
                    0.587148, 0.40736066666666665`, 0.09470066666666668], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    "RGBColor[0.880722, 0.611041, 0.142051]"], Appearance -> 
                    None, BaseStyle -> {}, BaselinePosition -> Baseline, 
                    DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.880722, 0.611041, 0.142051]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0.880722, 0.611041, 0.142051], Editable -> False,
                     Selectable -> False], ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   RowBox[{"Dashing", "[", 
                    RowBox[{"{", 
                    RowBox[{"Small", ",", "Small"}], "}"}], "]"}]}], "]"}], 
               ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.008333333333333333`", "]"}], 
                   ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0, 1, 0], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> RGBColor[0., 0.6666666666666666, 0.], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    "RGBColor[0, 1, 0]"], Appearance -> None, BaseStyle -> {},
                     BaselinePosition -> Baseline, DefaultBaseStyle -> {}, 
                    ButtonFunction :> With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0, 1, 0]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0, 1, 0], Editable -> False, Selectable -> 
                    False]}], "]"}], ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.008333333333333333`", "]"}], 
                   ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[1, 0, 0], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> RGBColor[0.6666666666666666, 0., 0.], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    "RGBColor[1, 0, 0]"], Appearance -> None, BaseStyle -> {},
                     BaselinePosition -> Baseline, DefaultBaseStyle -> {}, 
                    ButtonFunction :> With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[1, 0, 0]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[1, 0, 0], Editable -> False, Selectable -> 
                    False]}], "]"}], ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.008333333333333333`", "]"}], 
                   ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0, 0, 1], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> RGBColor[0., 0., 0.6666666666666666], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    "RGBColor[0, 0, 1]"], Appearance -> None, BaseStyle -> {},
                     BaselinePosition -> Baseline, DefaultBaseStyle -> {}, 
                    ButtonFunction :> With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0, 0, 1]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0, 0, 1], Editable -> False, Selectable -> 
                    False]}], "]"}]}], "}"}], ",", 
           RowBox[{"{", 
             RowBox[{#, ",", #2, ",", #3, ",", #4, ",", #5}], "}"}], ",", 
           RowBox[{"LegendMarkers", "\[Rule]", 
             RowBox[{"{", 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}], ",", 
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}], ",", 
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}], ",", 
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}], ",", 
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}]}], "}"}]}], ",", 
           RowBox[{"Joined", "\[Rule]", 
             RowBox[{"{", 
               
               RowBox[{
                "True", ",", "True", ",", "True", ",", "True", ",", "True"}], 
               "}"}]}], ",", 
           RowBox[{"LabelStyle", "\[Rule]", 
             RowBox[{"{", "}"}]}], ",", 
           RowBox[{"LegendLayout", "\[Rule]", "\"Column\""}]}], "]"}]& ), 
      Editable -> True], TraditionalForm], TraditionalForm]},
  "Legended",
  DisplayFunction->(GridBox[{{
      TagBox[
       ItemBox[
        PaneBox[
         TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
         BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
       "SkipImageSizeLevel"], 
      ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
    GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
    AutoDelete -> False, GridBoxItemSize -> Automatic, 
    BaselinePosition -> {1, 1}]& ),
  Editable->True,
  InterpretationFunction->(RowBox[{"Legended", "[", 
     RowBox[{#, ",", 
       RowBox[{"Placed", "[", 
         RowBox[{#2, ",", "After"}], "]"}]}], "]"}]& )]], "Output",
 CellChangeTimes->{{3.8086405575733833`*^9, 3.8086405759342995`*^9}, {
   3.808640689257425*^9, 3.8086407354943447`*^9}, {3.8086408209389257`*^9, 
   3.80864085661965*^9}, 3.808640908335397*^9, 3.808640946723781*^9, {
   3.808641005831541*^9, 3.808641013404298*^9}, 3.8086410587430906`*^9, 
   3.8086490687302475`*^9}],

Cell[BoxData[
 TemplateBox[{GraphicsBox[{{}, {{{}, {}, {
        Hue[0.67, 0.6, 0.6], 
        Directive[
         PointSize[0.006944444444444445], 
         AbsoluteThickness[1.6], 
         GrayLevel[0]], 
        LineBox[CompressedData["
1:eJxd0jmMVWUYh/HLohjqqTF2mNBZE+ykhXoqE22FFlqoaWi0sIHGZhITEyEh
RkXjEiDIdlEB586wzD6o99x9kXMe/f7xOc3JM8s5531/31vvf3Tyg72tVuvI
nlarvp9dWHqvc+bcsda/F32x9JVD5zdPXPlUv79c+pOP6+sz/f1S6a0T9U8+
1/9/Ufpg+9Kpowe/1POuln578cK31eFrev5XpS+/+u/2pa/1vm9Kv9Nc1/X+
70pfrw6/esP3+p4fSp9sXvijvu+n0p0z9QN/1vfeKH26/rxTN/X9t0pz3dY8
6ePNC37RfOn6bYsX7mjeu6Wbcdp3Nf+90m/Wn3PovvaR/rC5Hmg/7dL10xaW
2trXw9K9+nHVQ+3v19LNeo7+pn2mm887/7v2+6j0zRv19Uj7fly6GXfhifaf
bh63+Ic8ljXPsnyWNd+yvDqatyO/juZfkeeK9rEi31XtZ1Xeq9rXU/k/1f7S
nIdn2mea8/FM+32u8/Jc+05zfl5o/2nO0wt5rMljTR5r8liTx7o81uWxLo8N
eWzIY0Mem/LYlMemPLbksSWPLXlsy2NbHtvy2JHHjjx25LErj1157Mrjpeb5
r18ewyfNfGm80sybxi/N/H+WxjPNPtL4ptlPGu80+/qrNP5p9pfmPKTZZ5rz
kWa/f5fmvKTZd5rzk2b/ac5Tmnu3NB5deXTl0ZVHVx5deXTlUcmjkkclj0oe
lTwqefTk0ZNHTx49efTk0ZNHXx59efTl0ZdHXx59eQzkMZDHQB4DeQzkMZDH
QB5DeQzlMZTHUB5DeQzlMZLHSB4jeYzkMZLHSB5jeYzlMZbHWB5jeYzlMZHH
RB4TeUzkMZHHRB4TeUzlMZXHVB5TeUzlMZXHTB4zeczkMZPHTB4zeczlMZfH
XB5zeczlMZdH693/z5PGJ818abzSzJvGL838aTzT7CONb5r9pPFOs689pfFP
s7805yHNPtOcjzT7TXNe0uw7zflJs/805ynNfW9pPNJ4pPFI45HGI41HGo80
Hmk80nik8UjjkcZjX2k80nik8UjjkcYjjUcajzQeaTzSeKTxSHPfXxqPNB5p
PNJ4pPFI45HGI41HGo80Hmk80nik8XitNB5pPNJ4pPFI45HGI41HGo80Hmk8
0nikub9eGo80Hmk80nik8UjjkcYjjUcajzQeaTzSeKTxOFAajzQeaTzSeKTx
SOORxiONRxqPNB5pPNLc3yj9Dy9QTDs=
         "]]}, {
        Hue[0.9060679774997897, 0.6, 0.6], 
        Directive[
         PointSize[0.006944444444444445], 
         RGBColor[0.880722, 0.611041, 0.142051], 
         AbsoluteThickness[1.6], 
         Dashing[{Small, Small}]], 
        LineBox[CompressedData["
1:eJw9lXlcTdsfhveZyzUUmX6GZEiUropM6buMmULmITdDuBS5kRsKqQi3QaYS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         "]]}, {
        Hue[0.1421359549995791, 0.6, 0.6], 
        Directive[
         PointSize[0.006944444444444445], 
         AbsoluteThickness[1.6], 
         RGBColor[0, 1, 0]], 
        LineBox[CompressedData["
1:eJxF13k8ldkbAPD37rQvFNWPJNG0d1vsz8Gk1WSpmZiihbRNy0jaSIWxjKTR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         "]]}, {
        Hue[0.37820393249936934`, 0.6, 0.6], 
        Directive[
         PointSize[0.006944444444444445], 
         AbsoluteThickness[1.6], 
         RGBColor[1, 0, 0]], 
        LineBox[CompressedData["
1:eJxF1nlcVOUawPHZGBATcAlzBSsVFTHFNYTnNQFxvahkddFcgNxSQVNUvFcL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         "]]}, {
        Hue[0.6142719099991583, 0.6, 0.6], 
        Directive[
         PointSize[0.006944444444444445], 
         AbsoluteThickness[1.6], 
         RGBColor[0, 0, 1]], 
        LineBox[CompressedData["
1:eJxd1Hk81PkfB/D5zkVCu6VapO1YXSKSSMz7UyGktuhQsRnSri7SpexuW1vp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         "]]}}}, {}, {}, {{}, {}}}, {
    DisplayFunction -> Identity, PlotRangePadding -> {{
        Scaled[0.02], 
        Scaled[0.02]}, {
        Scaled[0.05], 
        Scaled[0.05]}}, AxesOrigin -> {0, 0}, 
     PlotRange -> {{0, 3.}, {-0.0912934135989736, 1.1689206388851379`}}, 
     PlotRangeClipping -> True, ImagePadding -> All, DisplayFunction -> 
     Identity, AspectRatio -> NCache[GoldenRatio^(-1), 0.6180339887498948], 
     Axes -> {True, True}, AxesLabel -> {None, None}, AxesOrigin -> {0, 0}, 
     DisplayFunction :> Identity, Frame -> {{False, False}, {False, False}}, 
     FrameLabel -> {{None, None}, {None, None}}, 
     FrameTicks -> {{Automatic, Automatic}, {Automatic, Automatic}}, 
     GridLines -> {None, None}, GridLinesStyle -> Directive[
       GrayLevel[0.5, 0.4]], ImageSize -> 700, 
     Method -> {"CoordinatesToolOptions" -> {"DisplayFunction" -> ({
           (Part[{{Identity, Identity}, {Identity, Identity}}, 1, 2][#]& )[
            Part[#, 1]], 
           (Part[{{Identity, Identity}, {Identity, Identity}}, 2, 2][#]& )[
            Part[#, 2]]}& ), "CopiedValueFunction" -> ({
           (Part[{{Identity, Identity}, {Identity, Identity}}, 1, 2][#]& )[
            Part[#, 1]], 
           (Part[{{Identity, Identity}, {Identity, Identity}}, 2, 2][#]& )[
            Part[#, 2]]}& )}}, 
     PlotRange -> {{0, 3.}, {-0.0912934135989736, 1.1689206388851379`}}, 
     PlotRangeClipping -> True, PlotRangePadding -> {{
        Scaled[0.02], 
        Scaled[0.02]}, {
        Scaled[0.05], 
        Scaled[0.05]}}, Ticks -> {Automatic, Automatic}}],FormBox[
    FormBox[
     TemplateBox[{
      "\"f(t)\"", "\"Gaver\"", "\"Euler\"", "\"CME\"", "\"CMER\""}, 
      "LineLegend", DisplayFunction -> (FormBox[
        StyleBox[
         StyleBox[
          PaneBox[
           TagBox[
            GridBox[{{
               TagBox[
                GridBox[{{
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    AbsoluteThickness[1.6], 
                    GrayLevel[0]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    AbsoluteThickness[1.6], 
                    GrayLevel[0]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    AbsoluteThickness[1.6], 
                    Dashing[{Small, Small}]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    AbsoluteThickness[1.6], 
                    Dashing[{Small, Small}]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #2}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0, 1, 0]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0, 1, 0]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #3}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    AbsoluteThickness[1.6], 
                    RGBColor[1, 0, 0]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    AbsoluteThickness[1.6], 
                    RGBColor[1, 0, 0]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #4}, {
                   GraphicsBox[{{
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0, 0, 1]], {
                    LineBox[{{0, 10}, {40, 10}}]}}, {
                    Directive[
                    EdgeForm[
                    Directive[
                    Opacity[0.3], 
                    GrayLevel[0]]], 
                    PointSize[0.0625], 
                    AbsoluteThickness[1.6], 
                    RGBColor[0, 0, 1]], {}}}, AspectRatio -> Full, 
                    ImageSize -> {40, 10}, PlotRangePadding -> None, 
                    ImagePadding -> Automatic, 
                    BaselinePosition -> (Scaled[0.1] -> Baseline)], #5}}, 
                 GridBoxAlignment -> {
                  "Columns" -> {Center, Left}, "Rows" -> {{Baseline}}}, 
                 AutoDelete -> False, 
                 GridBoxDividers -> {
                  "Columns" -> {{False}}, "Rows" -> {{False}}}, 
                 GridBoxItemSize -> {"Columns" -> {{All}}, "Rows" -> {{All}}},
                  GridBoxSpacings -> {
                  "Columns" -> {{0.5}}, "Rows" -> {{0.8}}}], "Grid"]}}, 
             GridBoxAlignment -> {"Columns" -> {{Left}}, "Rows" -> {{Top}}}, 
             AutoDelete -> False, 
             GridBoxItemSize -> {
              "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
             GridBoxSpacings -> {"Columns" -> {{1}}, "Rows" -> {{0}}}], 
            "Grid"], Alignment -> Left, AppearanceElements -> None, 
           ImageMargins -> {{5, 5}, {5, 5}}, ImageSizeAction -> 
           "ResizeToFit"], LineIndent -> 0, StripOnInput -> False], {
         FontFamily -> "Arial"}, Background -> Automatic, StripOnInput -> 
         False], TraditionalForm]& ), 
      InterpretationFunction :> (RowBox[{"LineLegend", "[", 
         RowBox[{
           RowBox[{"{", 
             RowBox[{
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.006944444444444445`", "]"}], 
                   ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    GrayLevel[0], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> GrayLevel[0.], FrameTicks -> None, 
                    PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], "GrayLevel[0]"], 
                    Appearance -> None, BaseStyle -> {}, BaselinePosition -> 
                    Baseline, DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    GrayLevel[0]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["GrayLevelColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    GrayLevel[0], Editable -> False, Selectable -> False]}], 
                 "]"}], ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.006944444444444445`", "]"}], 
                   ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0.880722, 0.611041, 0.142051], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> 
                    RGBColor[
                    0.587148, 0.40736066666666665`, 0.09470066666666668], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    "RGBColor[0.880722, 0.611041, 0.142051]"], Appearance -> 
                    None, BaseStyle -> {}, BaselinePosition -> Baseline, 
                    DefaultBaseStyle -> {}, ButtonFunction :> 
                    With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0.880722, 0.611041, 0.142051]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0.880722, 0.611041, 0.142051], Editable -> False,
                     Selectable -> False], ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   RowBox[{"Dashing", "[", 
                    RowBox[{"{", 
                    RowBox[{"Small", ",", "Small"}], "}"}], "]"}]}], "]"}], 
               ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.006944444444444445`", "]"}], 
                   ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0, 1, 0], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> RGBColor[0., 0.6666666666666666, 0.], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    "RGBColor[0, 1, 0]"], Appearance -> None, BaseStyle -> {},
                     BaselinePosition -> Baseline, DefaultBaseStyle -> {}, 
                    ButtonFunction :> With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0, 1, 0]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0, 1, 0], Editable -> False, Selectable -> 
                    False]}], "]"}], ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.006944444444444445`", "]"}], 
                   ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[1, 0, 0], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> RGBColor[0.6666666666666666, 0., 0.], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    "RGBColor[1, 0, 0]"], Appearance -> None, BaseStyle -> {},
                     BaselinePosition -> Baseline, DefaultBaseStyle -> {}, 
                    ButtonFunction :> With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[1, 0, 0]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[1, 0, 0], Editable -> False, Selectable -> 
                    False]}], "]"}], ",", 
               RowBox[{"Directive", "[", 
                 RowBox[{
                   RowBox[{"PointSize", "[", "0.006944444444444445`", "]"}], 
                   ",", 
                   RowBox[{"AbsoluteThickness", "[", "1.6`", "]"}], ",", 
                   InterpretationBox[
                    ButtonBox[
                    TooltipBox[
                    GraphicsBox[{{
                    GrayLevel[0], 
                    RectangleBox[{0, 0}]}, {
                    GrayLevel[0], 
                    RectangleBox[{1, -1}]}, {
                    RGBColor[0, 0, 1], 
                    RectangleBox[{0, -1}, {2, 1}]}}, AspectRatio -> 1, Frame -> 
                    True, FrameStyle -> RGBColor[0., 0., 0.6666666666666666], 
                    FrameTicks -> None, PlotRangePadding -> None, ImageSize -> 
                    Dynamic[{
                    Automatic, 1.35 CurrentValue["FontCapHeight"]/
                    AbsoluteCurrentValue[Magnification]}]], 
                    "RGBColor[0, 0, 1]"], Appearance -> None, BaseStyle -> {},
                     BaselinePosition -> Baseline, DefaultBaseStyle -> {}, 
                    ButtonFunction :> With[{Typeset`box$ = EvaluationBox[]}, 
                    If[
                    Not[
                    AbsoluteCurrentValue["Deployed"]], 
                    SelectionMove[Typeset`box$, All, Expression]; 
                    FrontEnd`Private`$ColorSelectorInitialAlpha = 1; 
                    FrontEnd`Private`$ColorSelectorInitialColor = 
                    RGBColor[0, 0, 1]; 
                    FrontEnd`Private`$ColorSelectorUseMakeBoxes = True; 
                    MathLink`CallFrontEnd[
                    FrontEnd`AttachCell[Typeset`box$, 
                    FrontEndResource["RGBColorValueSelector"], {
                    0, {Left, Bottom}}, {Left, Top}, 
                    "ClosingActions" -> {
                    "SelectionDeparture", "ParentChanged", 
                    "EvaluatorQuit"}]]]], BaseStyle -> Inherited, Evaluator -> 
                    Automatic, Method -> "Preemptive"], 
                    RGBColor[0, 0, 1], Editable -> False, Selectable -> 
                    False]}], "]"}]}], "}"}], ",", 
           RowBox[{"{", 
             RowBox[{#, ",", #2, ",", #3, ",", #4, ",", #5}], "}"}], ",", 
           RowBox[{"LegendMarkers", "\[Rule]", 
             RowBox[{"{", 
               RowBox[{
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}], ",", 
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}], ",", 
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}], ",", 
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}], ",", 
                 RowBox[{"{", 
                   RowBox[{"False", ",", "Automatic"}], "}"}]}], "}"}]}], ",", 
           RowBox[{"Joined", "\[Rule]", 
             RowBox[{"{", 
               
               RowBox[{
                "True", ",", "True", ",", "True", ",", "True", ",", "True"}], 
               "}"}]}], ",", 
           RowBox[{"LabelStyle", "\[Rule]", 
             RowBox[{"{", "}"}]}], ",", 
           RowBox[{"LegendLayout", "\[Rule]", "\"Column\""}]}], "]"}]& ), 
      Editable -> True], TraditionalForm], TraditionalForm]},
  "Legended",
  DisplayFunction->(GridBox[{{
      TagBox[
       ItemBox[
        PaneBox[
         TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
         BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
       "SkipImageSizeLevel"], 
      ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}}, 
    GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
    AutoDelete -> False, GridBoxItemSize -> Automatic, 
    BaselinePosition -> {1, 1}]& ),
  Editable->True,
  InterpretationFunction->(RowBox[{"Legended", "[", 
     RowBox[{#, ",", 
       RowBox[{"Placed", "[", 
         RowBox[{#2, ",", "After"}], "]"}]}], "]"}]& )]], "Output",
 CellChangeTimes->{{3.8086405575733833`*^9, 3.8086405759342995`*^9}, {
   3.808640689257425*^9, 3.8086407354943447`*^9}, {3.8086408209389257`*^9, 
   3.80864085661965*^9}, 3.808640908335397*^9, 3.808640946723781*^9, {
   3.808641005831541*^9, 3.808641013404298*^9}, 3.8086410587430906`*^9, 
   3.808649068913757*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"NILT", "[", 
   RowBox[{"Fexp", ",", "1", ",", "10", ",", "\"\<MExp\>\""}], "]"}], "//", 
  "N"}]], "Input",
 CellChangeTimes->{3.808565800076852*^9}],

Cell[BoxData["0.36874409352895937`"], "Output",
 CellChangeTimes->{3.8085658019294243`*^9, 3.8085764483838887`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["nParams"], "Input",
 CellChangeTimes->{{3.8085658040717306`*^9, 3.8085658056176567`*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "1", ",", "2", ",", "3", ",", "4", ",", "5", ",", "6", ",", "7", ",", "8", 
   ",", "9", ",", "10", ",", "11", ",", "12", ",", "13", ",", "14", ",", "15",
    ",", "16", ",", "17", ",", "18", ",", "19", ",", "20", ",", "21", ",", 
   "22", ",", "23", ",", "24", ",", "25", ",", "26", ",", "27", ",", "28", 
   ",", "29", ",", "30", ",", "31", ",", "32", ",", "33", ",", "34", ",", 
   "35", ",", "36", ",", "37", ",", "38", ",", "39", ",", "40", ",", "41", 
   ",", "42", ",", "43", ",", "44", ",", "45", ",", "46", ",", "47", ",", 
   "48", ",", "49", ",", "50", ",", "51", ",", "52", ",", "53", ",", "54", 
   ",", "55", ",", "56", ",", "57", ",", "58", ",", "59", ",", "60", ",", 
   "61", ",", "62", ",", "63", ",", "64", ",", "65", ",", "66", ",", "67", 
   ",", "68", ",", "69", ",", "70", ",", "71", ",", "72", ",", "73", ",", 
   "74", ",", "10", ",", "15", ",", "20", ",", "25", ",", "30", ",", "35", 
   ",", "40", ",", "45", ",", "50", ",", "55", ",", "60", ",", "65", ",", 
   "70", ",", "75", ",", "80", ",", "85", ",", "90", ",", "95", ",", "100", 
   ",", "105", ",", "110", ",", "115", ",", "120", ",", "125", ",", "130", 
   ",", "135", ",", "140", ",", "145", ",", "150", ",", "155", ",", "160", 
   ",", "165", ",", "170", ",", "175", ",", "180", ",", "185", ",", "190", 
   ",", "195", ",", "200", ",", "205", ",", "210", ",", "215", ",", "220", 
   ",", "225", ",", "230", ",", "235", ",", "240", ",", "245", ",", "250", 
   ",", "255", ",", "260", ",", "265", ",", "270", ",", "275", ",", "280", 
   ",", "285", ",", "290", ",", "295", ",", "300", ",", "305", ",", "310", 
   ",", "315", ",", "320", ",", "325", ",", "330", ",", "335", ",", "340", 
   ",", "345", ",", "350", ",", "355", ",", "360", ",", "365", ",", "370", 
   ",", "375", ",", "380", ",", "385", ",", "390", ",", "395", ",", "400", 
   ",", "405", ",", "420", ",", "440", ",", "460", ",", "480", ",", "500", 
   ",", "520", ",", "540", ",", "560", ",", "580", ",", "600", ",", "620", 
   ",", "640", ",", "660", ",", "680", ",", "700", ",", "720", ",", "740", 
   ",", "760", ",", "780", ",", "800", ",", "820", ",", "840", ",", "860", 
   ",", "880", ",", "900", ",", "920", ",", "940", ",", "960", ",", "980", 
   ",", "1000", ",", "190", ",", "195", ",", "200", ",", "205", ",", "210", 
   ",", "215", ",", "220", ",", "225", ",", "230", ",", "235", ",", "240", 
   ",", "245", ",", "250", ",", "255", ",", "260", ",", "265", ",", "270", 
   ",", "275", ",", "280", ",", "285", ",", "290", ",", "295", ",", "300", 
   ",", "305", ",", "310", ",", "315", ",", "320", ",", "325", ",", "330", 
   ",", "335", ",", "340", ",", "345", ",", "350", ",", "355", ",", "360", 
   ",", "365", ",", "370", ",", "375", ",", "380", ",", "385", ",", "390", 
   ",", "395", ",", "400", ",", "405", ",", "420", ",", "440", ",", "460", 
   ",", "480", ",", "500", ",", "520", ",", "540", ",", "560", ",", "580", 
   ",", "600", ",", "620", ",", "640", ",", "660", ",", "680", ",", "700", 
   ",", "720", ",", "740", ",", "760", ",", "780", ",", "800", ",", "820", 
   ",", "840", ",", "860", ",", "880", ",", "900", ",", "920", ",", "940", 
   ",", "960", ",", "980", ",", "1000", ",", "190", ",", "195", ",", "200", 
   ",", "205", ",", "210", ",", "215", ",", "220", ",", "225", ",", "230", 
   ",", "235", ",", "240", ",", "245", ",", "250", ",", "255", ",", "260", 
   ",", "265", ",", "270", ",", "275", ",", "280", ",", "285", ",", "290", 
   ",", "295", ",", "300", ",", "305", ",", "310", ",", "315", ",", "320", 
   ",", "325", ",", "330", ",", "335", ",", "340", ",", "345", ",", "350", 
   ",", "355", ",", "360", ",", "365", ",", "370", ",", "375", ",", "380", 
   ",", "385", ",", "390", ",", "395", ",", "400", ",", "405", ",", "420", 
   ",", "440", ",", "460", ",", "480", ",", "500", ",", "520", ",", "540", 
   ",", "560", ",", "580", ",", "600", ",", "620", ",", "640", ",", "660", 
   ",", "680", ",", "700", ",", "720", ",", "740", ",", "760", ",", "780", 
   ",", "800", ",", "820", ",", "840", ",", "860", ",", "880", ",", "900", 
   ",", "920", ",", "940", ",", "960", ",", "980", ",", "1000"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.80856580640552*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{"NILT", "[", 
  RowBox[{"LTFunc_", ",", "t_", ",", "order_", ",", "ILTmethod_"}], 
  "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", "\[IndentingNewLine]", "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}]], "Input",
 CellChangeTimes->{{3.808565152758746*^9, 3.808565153235482*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ILT", "[", 
    RowBox[{"fun_", ",", " ", "T_", ",", " ", "maxFnEvals_", ",", " ", 
     RowBox[{"method_:", "\"\<cme\>\""}], ",", 
     RowBox[{"precision_:", "MachinePrecision"}], ",", 
     RowBox[{"retprecloss_:", "False"}]}], "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
      "params", ",", "i", ",", "\[Eta]", ",", "\[Beta]", ",", "\[Xi]", ",", 
       "n", ",", "k", ",", "prec", ",", "res", ",", "precloss", ",", "eulera",
        ",", "actFnEvals"}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"Length", "[", 
           RowBox[{"Position", "[", 
            RowBox[{
             RowBox[{"Names", "[", "\"\<Global`*\>\"", "]"}], ",", 
             "\"\<cmeParams\>\""}], "]"}], "]"}], "\[Equal]", "0"}], " ", "||",
          " ", 
         RowBox[{
          RowBox[{"Length", "[", "cmeParams", "]"}], "\[Equal]", "0"}]}], ",",
         "\[IndentingNewLine]", 
        RowBox[{"(*", 
         RowBox[{
          RowBox[{"Print", "[", "\"\<CME parameters loaded\>\"", "]"}], ";"}],
          "*)"}], "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"cmeParams", " ", "=", " ", 
          RowBox[{"Import", "[", 
           RowBox[{"FileNameJoin", "[", 
            RowBox[{"{", 
             RowBox[{
              RowBox[{"NotebookDirectory", "[", "]"}], ",", 
              "\"\<iltcme.json\>\""}], "}"}], "]"}], "]"}]}], ";"}]}], 
       "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"Length", "[", 
           RowBox[{"Position", "[", 
            RowBox[{
             RowBox[{"Names", "[", "\"\<Global`*\>\"", "]"}], ",", 
             "\"\<cmeParamsb\>\""}], "]"}], "]"}], "\[Equal]", "0"}], " ", "||",
          " ", 
         RowBox[{
          RowBox[{"Length", "[", "cmeParamsb", "]"}], "\[Equal]", "0"}]}], 
        ",", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"cmeParamsb", " ", "=", " ", 
          RowBox[{"Import", "[", 
           RowBox[{"FileNameJoin", "[", 
            RowBox[{"{", 
             RowBox[{
              RowBox[{"NotebookDirectory", "[", "]"}], ",", 
              "\"\<iltcme_6_heur2.json\>\""}], "}"}], "]"}], "]"}]}], ";"}]}],
        "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"precision", "\[LessEqual]", "0"}], ",", 
        RowBox[{"prec", "=", "MachinePrecision"}], ",", 
        RowBox[{"prec", "=", "precision"}]}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"method", "\[Equal]", "\"\<cme\>\""}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"maxFnEvals", "<", "75"}], ",", "\[IndentingNewLine]", 
           "\t", 
           RowBox[{
            RowBox[{"params", " ", "=", " ", 
             RowBox[{"cmeParams", "[", 
              RowBox[{"[", "1", "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
            "\t", 
            RowBox[{"For", "[", 
             RowBox[{
              RowBox[{"i", "=", "2"}], ",", 
              RowBox[{"i", "<", 
               RowBox[{"Length", "[", "cmeParams", "]"}]}], ",", 
              RowBox[{"i", "++"}], ",", "\[IndentingNewLine]", "\t", 
              RowBox[{"If", "[", 
               RowBox[{
                RowBox[{
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{"\"\<cv2\>\"", "/.", 
                    RowBox[{"cmeParams", "[", 
                    RowBox[{"[", "i", "]"}], "]"}]}], ")"}], " ", "<", 
                  RowBox[{"(", 
                   RowBox[{"\"\<cv2\>\"", "/.", "params"}], ")"}]}], " ", "&&",
                  " ", 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"\"\<n\>\"", "+", "1"}], " ", "/.", 
                    RowBox[{"cmeParams", "[", 
                    RowBox[{"[", "i", "]"}], "]"}]}], ")"}], "\[LessEqual]", 
                  "maxFnEvals"}]}], ",", 
                RowBox[{"params", "=", 
                 RowBox[{"cmeParams", "[", 
                  RowBox[{"[", "i", "]"}], "]"}]}]}], " ", "]"}]}], 
             "\[IndentingNewLine]", "\t", "]"}], ";", "\[IndentingNewLine]", 
            "\t", 
            RowBox[{"\[Eta]", "=", 
             RowBox[{"SetPrecision", "[", 
              RowBox[{
               RowBox[{"Prepend", "[", 
                RowBox[{
                 RowBox[{
                  RowBox[{
                   RowBox[{"\"\<a\>\"", "*", "\"\<mu1\>\""}], "+", 
                   RowBox[{
                   "\[ImaginaryI]", "*", "\"\<b\>\"", "*", "\"\<mu1\>\""}]}], 
                  "/.", "params"}], ",", 
                 RowBox[{
                  RowBox[{"\"\<c\>\"", "*", "\"\<mu1\>\""}], "/.", 
                  "params"}]}], "]"}], ",", "prec"}], "]"}]}], ";", 
            "\[IndentingNewLine]", "\t", 
            RowBox[{"\[Beta]", "=", 
             RowBox[{"SetPrecision", "[", 
              RowBox[{
               RowBox[{
                RowBox[{
                 RowBox[{"Prepend", "[", 
                  RowBox[{
                   RowBox[{"1", "+", 
                    RowBox[{
                    "\[ImaginaryI]", " ", "*", " ", "\"\<omega\>\"", "*", 
                    RowBox[{"Range", "[", 
                    RowBox[{"\"\<n\>\"", "/.", "params"}], "]"}]}]}], ",", 
                   "1"}], "]"}], "*", "\"\<mu1\>\""}], " ", "/.", "params"}], 
               ",", "prec"}], "]"}]}], ";"}], "\[IndentingNewLine]", "\t", 
           ",", "\[IndentingNewLine]", "\t", 
           RowBox[{
            RowBox[{"params", " ", "=", " ", 
             RowBox[{"cmeParamsb", "[", 
              RowBox[{"[", "1", "]"}], "]"}]}], ";", "\[IndentingNewLine]", 
            "\t", 
            RowBox[{"For", "[", 
             RowBox[{
              RowBox[{"i", "=", "2"}], ",", 
              RowBox[{"i", "<", 
               RowBox[{"Length", "[", "cmeParamsb", "]"}]}], ",", 
              RowBox[{"i", "++"}], ",", "\[IndentingNewLine]", "\t", 
              RowBox[{"If", "[", 
               RowBox[{
                RowBox[{
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{"\"\<cv2\>\"", "/.", 
                    RowBox[{"cmeParamsb", "[", 
                    RowBox[{"[", "i", "]"}], "]"}]}], ")"}], " ", "<", 
                  RowBox[{"(", 
                   RowBox[{"\"\<cv2\>\"", "/.", "params"}], ")"}]}], " ", "&&",
                  " ", 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{
                    RowBox[{"\"\<n\>\"", "+", "1"}], " ", "/.", 
                    RowBox[{"cmeParamsb", "[", 
                    RowBox[{"[", "i", "]"}], "]"}]}], ")"}], "\[LessEqual]", 
                  "maxFnEvals"}]}], ",", 
                RowBox[{"params", "=", 
                 RowBox[{"cmeParamsb", "[", 
                  RowBox[{"[", "i", "]"}], "]"}]}]}], " ", "]"}]}], 
             "\[IndentingNewLine]", "\t", "]"}], ";", "\[IndentingNewLine]", 
            "\t", 
            RowBox[{"\[Eta]", "=", 
             RowBox[{"SetPrecision", "[", 
              RowBox[{
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"\"\<eta_re\>\"", "+", 
                  RowBox[{"\[ImaginaryI]", "*", "\"\<eta_im\>\""}]}], ")"}], "/.",
                 "params"}], ",", "prec"}], "]"}]}], ";", 
            "\[IndentingNewLine]", "\t", 
            RowBox[{"actFnEvals", "=", 
             RowBox[{"Length", "[", "\[Eta]", "]"}]}], ";", 
            "\[IndentingNewLine]", "\t", 
            RowBox[{"\[Beta]", "=", 
             RowBox[{"SetPrecision", "[", 
              RowBox[{
               RowBox[{"Flatten", "[", 
                RowBox[{"Table", "[", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"(", 
                    RowBox[{"\"\<beta_re\>\"", "+", 
                    RowBox[{
                    "\[ImaginaryI]", "*", "\"\<\[CapitalDelta]beta_im\>\"", 
                    "*", "k"}]}], ")"}], "/.", "params"}], ",", 
                  RowBox[{"{", 
                   RowBox[{"k", ",", "0", ",", 
                    RowBox[{"actFnEvals", "-", "1"}]}], "}"}]}], "]"}], "]"}],
                ",", "prec"}], "]"}]}], ";"}]}], "\[IndentingNewLine]", "]"}],
          ";"}], "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"method", "\[Equal]", "\"\<euler\>\""}], ",", 
           "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"n", "=", 
             RowBox[{"If", "[", 
              RowBox[{
               RowBox[{
                RowBox[{"Mod", "[", 
                 RowBox[{"maxFnEvals", ",", "2"}], "]"}], "\[Equal]", "0"}], 
               ",", 
               RowBox[{"maxFnEvals", "-", "1"}], ",", "maxFnEvals"}], "]"}]}],
             ";", "\[IndentingNewLine]", 
            RowBox[{"\[Beta]", "=", 
             RowBox[{"Table", "[", 
              RowBox[{
               RowBox[{"SetPrecision", "[", 
                RowBox[{
                 RowBox[{
                  FractionBox[
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{"n", "-", "1"}], ")"}], 
                    RowBox[{"Log", "[", "10", "]"}]}], "6"], "+", 
                  RowBox[{"\[Pi]", " ", "\[ImaginaryI]", " ", 
                   RowBox[{"(", 
                    RowBox[{"k", "-", "1"}], ")"}]}]}], ",", "prec"}], "]"}], 
               ",", 
               RowBox[{"{", 
                RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
            "\[IndentingNewLine]", 
            RowBox[{"\[Xi]", "=", 
             RowBox[{"ConstantArray", "[", 
              RowBox[{
               RowBox[{"SetPrecision", "[", 
                RowBox[{"1", ",", "prec"}], "]"}], ",", "n"}], "]"}]}], ";", 
            "\[IndentingNewLine]", 
            RowBox[{
             RowBox[{"\[Xi]", "[", 
              RowBox[{"[", "1", "]"}], "]"}], " ", "=", " ", 
             RowBox[{"SetPrecision", "[", 
              RowBox[{
               RowBox[{"1", "/", "2"}], ",", "prec"}], "]"}]}], ";", 
            "\[IndentingNewLine]", 
            RowBox[{
             RowBox[{"\[Xi]", "[", 
              RowBox[{"[", "n", "]"}], "]"}], "=", 
             RowBox[{"SetPrecision", "[", 
              RowBox[{
               FractionBox["1", 
                SuperscriptBox["2", 
                 RowBox[{
                  RowBox[{"(", 
                   RowBox[{"n", "-", "1"}], ")"}], "/", "2"}]]], ",", 
               "prec"}], "]"}]}], ";", "\[IndentingNewLine]", 
            RowBox[{"For", "[", 
             RowBox[{
              RowBox[{"k", "=", "1"}], ",", 
              RowBox[{"k", "<", 
               RowBox[{
                RowBox[{"(", 
                 RowBox[{"n", "-", "1"}], ")"}], "/", "2"}]}], ",", 
              RowBox[{"k", "++"}], ",", 
              RowBox[{
               RowBox[{"\[Xi]", "[", 
                RowBox[{"[", 
                 RowBox[{"n", "-", "k"}], "]"}], "]"}], "=", 
               RowBox[{
                RowBox[{"\[Xi]", "[", 
                 RowBox[{"[", 
                  RowBox[{"n", "-", "k", "+", "1"}], "]"}], "]"}], "+", " ", 
                RowBox[{
                 SuperscriptBox["2", 
                  RowBox[{
                   RowBox[{"-", 
                    RowBox[{"(", 
                    RowBox[{"n", "-", "1"}], ")"}]}], "/", "2"}]], 
                 RowBox[{"Binomial", "[", 
                  RowBox[{
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{"n", "-", "1"}], ")"}], "/", "2"}], ",", "k"}], 
                  "]"}]}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
            RowBox[{"\[Eta]", "=", 
             RowBox[{"Table", "[", 
              RowBox[{
               RowBox[{"SetPrecision", "[", 
                RowBox[{
                 RowBox[{
                  SuperscriptBox["10", 
                   RowBox[{
                    RowBox[{"(", 
                    RowBox[{"n", "-", "1"}], ")"}], "/", "6"}]], 
                  SuperscriptBox[
                   RowBox[{"(", 
                    RowBox[{"-", "1"}], ")"}], 
                   RowBox[{"k", "-", "1"}]], 
                  RowBox[{"\[Xi]", "[", 
                   RowBox[{"[", "k", "]"}], "]"}]}], ",", "prec"}], "]"}], 
               ",", 
               RowBox[{"{", 
                RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
            "\[IndentingNewLine]", 
            RowBox[{"\[Eta]", "=", 
             RowBox[{"Table", "[", 
              RowBox[{
               RowBox[{
                RowBox[{"SetPrecision", "[", 
                 RowBox[{
                  RowBox[{
                   SuperscriptBox["10", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{"n", "-", "1"}], ")"}], "/", "6"}]], 
                   SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{"-", "1"}], ")"}], 
                    RowBox[{"k", "-", "1"}]]}], ",", "prec"}], "]"}], "*", 
                RowBox[{"\[Xi]", "[", 
                 RowBox[{"[", "k", "]"}], "]"}]}], ",", 
               RowBox[{"{", 
                RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";"}], 
           "\[IndentingNewLine]", ",", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"If", "[", 
             RowBox[{
              RowBox[{"method", "\[Equal]", "\"\<gaver\>\""}], ",", 
              "\[IndentingNewLine]", 
              RowBox[{
               RowBox[{"n", "=", 
                RowBox[{"If", "[", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"Mod", "[", 
                    RowBox[{"maxFnEvals", ",", "2"}], "]"}], "\[Equal]", 
                   "1"}], ",", 
                  RowBox[{"maxFnEvals", "-", "1"}], ",", "maxFnEvals"}], 
                 "]"}]}], ";", "\[IndentingNewLine]", 
               RowBox[{"\[Beta]", "=", 
                RowBox[{"Table", "[", 
                 RowBox[{
                  RowBox[{"k", " ", 
                   RowBox[{"SetPrecision", "[", 
                    RowBox[{
                    RowBox[{"Log", "[", "2", "]"}], ",", "prec"}], "]"}]}], 
                  ",", 
                  RowBox[{"{", 
                   RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
               "\[IndentingNewLine]", 
               RowBox[{"\[Eta]", "=", 
                RowBox[{"Table", "[", 
                 RowBox[{
                  RowBox[{
                   RowBox[{"SetPrecision", "[", 
                    RowBox[{
                    RowBox[{"Log", "[", "2", "]"}], ",", "prec"}], "]"}], 
                   SuperscriptBox[
                    RowBox[{"(", 
                    RowBox[{"-", "1"}], ")"}], 
                    RowBox[{
                    RowBox[{"n", "/", "2"}], "+", "k"}]], 
                   RowBox[{"Sum", "[", 
                    RowBox[{
                    RowBox[{
                    FractionBox[
                    SuperscriptBox["j", 
                    RowBox[{
                    RowBox[{"n", "/", "2"}], "+", "1"}]], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{"n", "/", "2"}], ")"}], "!"}]], 
                    RowBox[{"Binomial", "[", 
                    RowBox[{
                    RowBox[{"n", "/", "2"}], ",", "j"}], "]"}], 
                    RowBox[{"Binomial", "[", 
                    RowBox[{
                    RowBox[{"2", "j"}], ",", "j"}], "]"}], 
                    RowBox[{"Binomial", "[", 
                    RowBox[{"j", ",", 
                    RowBox[{"k", "-", "j"}]}], "]"}]}], ",", 
                    RowBox[{"{", 
                    RowBox[{"j", ",", 
                    RowBox[{"Floor", "[", 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{"k", "+", "1"}], ")"}], "/", "2"}], "]"}], ",", 
                    RowBox[{"Min", "[", 
                    RowBox[{"k", ",", 
                    RowBox[{"n", "/", "2"}]}], "]"}]}], "}"}]}], "]"}]}], 
                  "  ", ",", 
                  RowBox[{"{", 
                   RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], 
               ";"}]}], "\[IndentingNewLine]", "]"}], ";"}]}], 
          "\[IndentingNewLine]", "]"}], ";"}]}], "\[IndentingNewLine]", "]"}],
       ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"method", " ", "\[Equal]", " ", "\"\<euler2\>\""}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"n", "=", 
          RowBox[{"If", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"Mod", "[", 
              RowBox[{"maxFnEvals", ",", "2"}], "]"}], "\[Equal]", "0"}], ",", 
            RowBox[{"maxFnEvals", "-", "1"}], ",", "maxFnEvals"}], "]"}]}], 
         ";", "\[IndentingNewLine]", 
         RowBox[{"eulera", "=", 
          RowBox[{
           RowBox[{"13", "/", "12"}], "*", 
           SuperscriptBox[
            RowBox[{"Log", "[", "n", "]"}], 
            RowBox[{"2", "/", "3"}]]}]}], ";", "\[IndentingNewLine]", 
         RowBox[{"\[Beta]", "=", 
          RowBox[{"Table", "[", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{"eulera", "+", 
              RowBox[{"\[Pi]", 
               RowBox[{"(", " ", 
                RowBox[{"k", "-", "1"}], ")"}], " ", "\[ImaginaryI]"}]}], 
             ")"}], ",", 
            RowBox[{"{", 
             RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"\[Xi]", "=", 
          RowBox[{"ConstantArray", "[", 
           RowBox[{
            RowBox[{"SetPrecision", "[", 
             RowBox[{"1", ",", "prec"}], "]"}], ",", "n"}], "]"}]}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"\[Xi]", "[", 
           RowBox[{"[", "1", "]"}], "]"}], " ", "=", " ", 
          RowBox[{"SetPrecision", "[", 
           RowBox[{
            RowBox[{"1", "/", "2"}], ",", "prec"}], "]"}]}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"\[Xi]", "[", 
           RowBox[{"[", "n", "]"}], "]"}], "=", 
          RowBox[{"SetPrecision", "[", 
           RowBox[{
            FractionBox["1", 
             SuperscriptBox["2", 
              RowBox[{
               RowBox[{"(", 
                RowBox[{"n", "-", "1"}], ")"}], "/", "2"}]]], ",", "prec"}], 
           "]"}]}], ";", "\[IndentingNewLine]", 
         RowBox[{"For", "[", 
          RowBox[{
           RowBox[{"k", "=", "1"}], ",", 
           RowBox[{"k", "<", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"n", "-", "1"}], ")"}], "/", "2"}]}], ",", 
           RowBox[{"k", "++"}], ",", 
           RowBox[{
            RowBox[{"\[Xi]", "[", 
             RowBox[{"[", 
              RowBox[{"n", "-", "k"}], "]"}], "]"}], "=", 
            RowBox[{
             RowBox[{"\[Xi]", "[", 
              RowBox[{"[", 
               RowBox[{"n", "-", "k", "+", "1"}], "]"}], "]"}], "+", " ", 
             RowBox[{
              SuperscriptBox["2", 
               RowBox[{
                RowBox[{"-", 
                 RowBox[{"(", 
                  RowBox[{"n", "-", "1"}], ")"}]}], "/", "2"}]], 
              RowBox[{"Binomial", "[", 
               RowBox[{
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{"n", "-", "1"}], ")"}], "/", "2"}], ",", "k"}], 
               "]"}]}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
         "\[IndentingNewLine]", 
         RowBox[{"\[Eta]", "=", 
          RowBox[{"Table", "[", 
           RowBox[{
            RowBox[{"SetPrecision", "[", 
             RowBox[{
              RowBox[{
               SuperscriptBox["\[ExponentialE]", "eulera"], 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{"-", "1"}], ")"}], 
                RowBox[{"k", "-", "1"}]], 
               RowBox[{"\[Xi]", "[", 
                RowBox[{"[", "k", "]"}], "]"}]}], ",", "prec"}], "]"}], ",", 
            RowBox[{"{", 
             RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";"}]}], 
       "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"method", "\[Equal]", "\"\<cmer\>\""}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{
          RowBox[{"{", 
           RowBox[{"\[Eta]", ",", "\[Beta]"}], "}"}], "=", 
          RowBox[{"CMER", "[", 
           RowBox[{"Ceiling", "[", 
            RowBox[{"maxFnEvals", "/", "2"}], "]"}], "]"}]}], ";"}]}], 
       "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"res", "=", 
       RowBox[{
        RowBox[{"Re", "[", 
         RowBox[{
          RowBox[{"Sum", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"\[Eta]", "[", 
              RowBox[{"[", "k", "]"}], "]"}], "*", 
             RowBox[{"fun", "[", 
              RowBox[{
               RowBox[{"\[Beta]", "[", 
                RowBox[{"[", "k", "]"}], "]"}], "/", "x"}], "]"}]}], ",", 
            RowBox[{"{", 
             RowBox[{"k", ",", "1", ",", 
              RowBox[{"Length", "[", "\[Eta]", "]"}]}], "}"}]}], "]"}], "/", 
          "x"}], "]"}], "/.", 
        RowBox[{"x", "\[Rule]", "T"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{"retprecloss", ",", "\[IndentingNewLine]", "\t", 
        RowBox[{
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"method", "\[Equal]", "\"\<cmer\>\""}], ",", 
           "\[IndentingNewLine]", "\t\t", 
           RowBox[{
            RowBox[{"precloss", "=", 
             RowBox[{
              RowBox[{"Precision", "[", "\[Eta]", "]"}], "-", 
              RowBox[{"Precision", "[", "res", "]"}]}]}], ";"}], 
           "\[IndentingNewLine]", "\t", ",", "\[IndentingNewLine]", "\t\t", 
           RowBox[{
            RowBox[{"precloss", "=", 
             RowBox[{"prec", "-", 
              RowBox[{"Precision", "[", "res", "]"}]}]}], ";"}]}], 
          "\[IndentingNewLine]", "\t", "]"}], ";", "\[IndentingNewLine]", 
         "\t", 
         RowBox[{"Return", "[", 
          RowBox[{"{", 
           RowBox[{"res", ",", "precloss"}], "}"}], "]"}], ";"}], 
        "\[IndentingNewLine]", ",", "\[IndentingNewLine]", "\t", 
        RowBox[{
         RowBox[{"Return", "[", "res", "]"}], ";"}]}], "\[IndentingNewLine]", 
       "]"}], ";"}]}], "\[IndentingNewLine]", "]"}]}], ";"}]], "Input"]
},
WindowSize->{1664, 967},
WindowMargins->{{Automatic, 32}, {Automatic, 0}},
FrontEndVersion->"11.0 for Microsoft Windows (64-bit) (July 28, 2016)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[558, 20, 23508, 583, 1705, "Input"],
Cell[24069, 605, 426, 6, 30, "Input"],
Cell[24498, 613, 1185, 34, 159, "Input"],
Cell[25686, 649, 4870, 144, 335, "Input"],
Cell[CellGroupData[{
Cell[30581, 797, 1595, 39, 50, "Input"],
Cell[32179, 838, 44370, 830, 455, "Output"],
Cell[76552, 1670, 40438, 764, 448, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[117027, 2439, 185, 5, 30, "Input"],
Cell[117215, 2446, 115, 1, 30, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[117367, 2452, 103, 1, 30, "Input"],
Cell[117473, 2455, 4198, 58, 145, "Output"]
}, Open  ]],
Cell[121686, 2516, 124, 3, 30, "Input"],
Cell[121813, 2521, 195, 3, 107, "Input"],
Cell[122011, 2526, 22973, 560, 1621, "Input"]
}
]
*)

